from __future__ import annotations

from typing import Any

from pydantic import BaseModel, field_validator

from bfabric_app_runner.specs.app.commands_spec import CommandsSpec  # noqa: TCH001
from bfabric_app_runner.specs.config_interpolation import interpolate_config_strings, VariablesApp


class AppVersion(BaseModel):
    """A concrete app version specification.

    For a better separation of concerns, the submitter will not be resolved automatically.
    """

    version: str = "latest"
    commands: CommandsSpec
    # TODO remove when new submitter becomes available
    reuse_default_resource: bool = True


class AppVersionTemplate(BaseModel):
    version: str
    commands: CommandsSpec
    # TODO remove when new submitter becomes available
    reuse_default_resource: bool = True

    def evaluate(self, variables_app: VariablesApp) -> AppVersion:
        """Evaluates the template to a concrete ``AppVersion`` instance."""
        data_template = self.model_dump(mode="json")
        data = interpolate_config_strings(data_template, variables={"app": variables_app, "workunit": None})
        return AppVersion.model_validate(data)


class AppVersionMultiTemplate(BaseModel):
    version: list[str]
    commands: CommandsSpec

    # TODO remove when new submitter becomes available
    reuse_default_resource: bool = True

    @field_validator("version", mode="before")
    def _version_ensure_list(cls, values: Any) -> list[str]:
        if not isinstance(values, list):
            return [values]
        return values

    def expand_versions(self) -> list[AppVersionTemplate]:
        """Returns a list of individual ``AppVersionTemplate`` instances, expanding each template of multiple versions.
        If substitutions are used they will not be expanded yet but rather when converting the template to a concrete
        AppVersion.
        """
        versions = []
        for version in self.version:
            version_data = self.model_dump(mode="json")
            version_data["version"] = version
            versions.append(AppVersionTemplate.model_validate(version_data))
        return versions



from __future__ import annotations

import hashlib
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path


def md5sum(file: Path) -> str:
    """Calculates the MD5 checksum of a file."""
    hasher = hashlib.md5()
    with file.open("rb") as f:
        for chunk in iter(lambda: f.read(16384), b""):
            hasher.update(chunk)
    return hasher.hexdigest()



from bfabric_app_runner.commands.command_exec import execute_command_exec
from bfabric_app_runner.specs.app.commands_spec import CommandShell, CommandExec


def execute_command_shell(command: CommandShell, *args: str) -> None:
    """Executes the command with the provided arguments."""
    execute_command_exec(CommandExec(command=command.command), *args)



import fcntl
import hashlib
import os
import shlex
import shutil
import socket
import tempfile
from pathlib import Path
from typing import Protocol, Annotated

import cyclopts
from loguru import logger

from bfabric_app_runner.commands.command_exec import execute_command_exec
from bfabric_app_runner.specs.app.commands_spec import CommandPythonEnv, CommandExec


class PythonEnvironment:
    """Represents a Python virtual environment with provisioning and execution capabilities."""

    def __init__(self, env_path: Path, command: CommandPythonEnv) -> None:
        self.env_path = env_path
        self.command = command
        self.python_executable = env_path / "bin" / "python"
        self.bin_path = env_path / "bin"

    def provision(self) -> None:
        """Provision the Python environment."""
        self._create_virtual_environment()
        self._install_dependencies()
        self._install_local_deps()
        self._mark_provisioned()

    @property
    def is_provisioned(self) -> bool:
        """True if the environment is already provisioned."""
        return self._provisioned_marker.exists()

    def _create_virtual_environment(self) -> None:
        """Create a virtual environment using uv venv."""
        self.env_path.parent.mkdir(parents=True, exist_ok=True)
        venv_cmd = ["uv", "venv", "-p", self.command.python_version, str(self.env_path)]
        exec_command = CommandExec(
            command=shlex.join(venv_cmd), env=self.command.env, prepend_paths=self.command.prepend_paths
        )
        execute_command_exec(exec_command)

    def _install_dependencies(self) -> None:
        """Install dependencies from pylock file."""
        install_cmd = ["uv", "pip", "install", "-p", str(self.python_executable), "-r", str(self.command.pylock)]
        if self.command.refresh:
            install_cmd.append("--reinstall")

        exec_command = CommandExec(
            command=shlex.join(install_cmd), env=self.command.env, prepend_paths=self.command.prepend_paths
        )
        execute_command_exec(exec_command)

    def _install_local_deps(self) -> None:
        """Install local extra dependencies with --no-deps."""
        if not self.command.local_extra_deps:
            return

        dep_install_cmd = ["uv", "pip", "install", "-p", str(self.python_executable), "--no-deps"]
        dep_install_cmd.extend(str(dep.absolute()) for dep in self.command.local_extra_deps)
        exec_command = CommandExec(
            command=shlex.join(dep_install_cmd), env=self.command.env, prepend_paths=self.command.prepend_paths
        )
        execute_command_exec(exec_command)

    def _mark_provisioned(self) -> None:
        """Mark environment as successfully provisioned."""
        self._provisioned_marker.touch()

    @property
    def _provisioned_marker(self) -> Path:
        """Path to the provisioned marker file."""
        return self.env_path / ".provisioned"


class EnvironmentStrategy(Protocol):
    """Protocol for obtaining Python environments."""

    def get_environment(self, command: CommandPythonEnv) -> PythonEnvironment:
        """Get a Python environment for the given command."""
        ...


class CachedEnvironmentStrategy:
    """Strategy for cached environments with file locking."""

    def get_environment(self, command: CommandPythonEnv) -> PythonEnvironment:
        env_path = self._get_cache_path(command)
        environment = PythonEnvironment(env_path, command)

        # Use file lock to prevent race conditions during provisioning
        lock_file = env_path.parent / f"{env_path.name}.lock"
        env_path.parent.mkdir(parents=True, exist_ok=True)

        with lock_file.open("a") as lock:
            fcntl.flock(lock.fileno(), fcntl.LOCK_EX)

            # Check if environment is properly provisioned (under lock)
            if not environment.is_provisioned:
                logger.info(f"Provisioning Python environment at {env_path} with command: {command.command}")
                environment.provision()

        return environment

    def _get_cache_path(self, command: CommandPythonEnv) -> Path:
        """Get the app_runner cache directory for Python environments."""
        cache_dir = Path(os.environ.get("XDG_CACHE_HOME", "~/.cache")).expanduser()
        app_runner_cache = cache_dir / "bfabric_app_runner" / "envs"
        env_hash = self._compute_env_hash(command)
        return app_runner_cache / env_hash

    def _compute_env_hash(self, command: CommandPythonEnv) -> str:
        """Returns a hash for the environment based on the command's specifications."""
        hostname = socket.gethostname()
        hash_input = f"{hostname}:{command.python_version}:{command.pylock.absolute()}:{command.pylock.stat().st_mtime}"
        if command.local_extra_deps:
            deps_str = ",".join(str(p.absolute()) for p in command.local_extra_deps)
            hash_input += f":{deps_str}"
        env_hash = hashlib.sha256(hash_input.encode()).hexdigest()[:16]
        logger.debug(f"Environment hash for input: {hash_input!r} is {env_hash!r}")
        return env_hash


class EphemeralEnvironmentStrategy:
    """Strategy for ephemeral (temporary) environments."""

    def get_environment(self, command: CommandPythonEnv) -> PythonEnvironment:
        env_path = self._get_ephemeral_path()
        environment = PythonEnvironment(env_path, command)

        logger.info(
            f"Provisioning ephemeral Python environment at {env_path} with command: "
            f"{command.command} (refresh mode)"
        )
        environment.provision()
        return environment

    def _get_ephemeral_path(self) -> Path:
        """Get an ephemeral directory for ephemeral (refresh) environments using cache directory."""
        # Use the same cache directory as cached environments
        cache_dir = Path(os.environ.get("XDG_CACHE_HOME", "~/.cache")).expanduser()
        app_runner_cache = cache_dir / "bfabric_app_runner" / "ephemeral"

        # Create a unique temporary subdirectory within the cache
        app_runner_cache.mkdir(parents=True, exist_ok=True)
        temp_dir = tempfile.mkdtemp(prefix="env_", dir=app_runner_cache)
        return Path(temp_dir)


def _ensure_environment(command: CommandPythonEnv) -> Path:
    """Ensure the Python environment exists and return its path."""
    strategy = EphemeralEnvironmentStrategy() if command.refresh else CachedEnvironmentStrategy()
    environment = strategy.get_environment(command)
    return environment.env_path


def execute_command_python_env(command: Annotated[CommandPythonEnv, cyclopts.Parameter(name="*")], *args: str) -> None:
    """Executes the command with the provided arguments using a Python environment."""
    # Ensure the environment exists (either cached or ephemeral)
    env_path = _ensure_environment(command)

    try:
        # Build command using the environment's Python interpreter
        command_args = shlex.split(command.command) + list(args)
        final_command = _ensure_executable(command_args, env_path)

        # Include venv's bin directory in prepend_paths
        venv_bin_path = env_path / "bin"
        prepend_paths = [venv_bin_path] + (command.prepend_paths or [])

        exec_command = CommandExec(command=shlex.join(final_command), env=command.env, prepend_paths=prepend_paths)
        execute_command_exec(exec_command)
    finally:
        # Clean up ephemeral environments after use
        if command.refresh:
            _cleanup_ephemeral_environment(env_path)


def _ensure_executable(command_args: list[str], env_path: Path) -> list[str]:
    candidate_script = env_path / "bin" / command_args[0]
    if candidate_script.exists():
        return [str(candidate_script), *command_args[1:]]

    # Fallback to using the Python interpreter directly
    python_executable = env_path / "bin" / "python"
    return [str(python_executable), *command_args]


def _cleanup_ephemeral_environment(env_path: Path) -> None:
    """Clean up ephemeral environment after use."""
    try:
        if env_path.exists():
            logger.info(f"Cleaning up ephemeral environment at {env_path}")
            shutil.rmtree(env_path)
    except (OSError, PermissionError) as e:
        logger.warning(f"Failed to cleanup ephemeral environment at {env_path}: {e}")


app = cyclopts.App()
app.command(execute_command_python_env, name="execute")

if __name__ == "__main__":
    app()



from __future__ import annotations

from typing import TYPE_CHECKING

from loguru import logger

if TYPE_CHECKING:
    from bfabric_app_runner.inputs.resolve.resolved_inputs import ResolvedStaticFile
    from pathlib import Path


def prepare_resolved_static_file(file: ResolvedStaticFile, working_dir: Path) -> None:
    """Prepares the file specified by the spec.

    This operation not modify the file and thus preserve the modification time, if the file already exists
    and has the same content.
    :param file: the resolved static file
    :param working_dir: the working directory
    :raises ValueError: if the path already exists but is not a file
    """
    path = working_dir / file.filename
    path.parent.mkdir(exist_ok=True, parents=True)
    _write_file_if_changed(content=file.content, path=path)


def _write_file_if_changed(content: str | bytes, path: Path) -> None:
    binary_flag = "b" if isinstance(content, bytes) else ""

    # check if the file already exists and has the same content
    if path.exists():
        if not path.is_file():
            msg = f"Path {path} exists but is not a file"
            raise ValueError(msg)
        with path.open(f"r{binary_flag}") as f:
            existing_content = f.read()
        if existing_content == content:
            logger.debug(f"Skipping {path} as it already exists and has the same content")
            return

    # write the file
    with path.open(f"w{binary_flag}") as f:
        f.write(content)
    logger.info(f"Writen to {path}")



from __future__ import annotations

from typing import TYPE_CHECKING, assert_never

from bfabric_app_runner.inputs.resolve.resolved_inputs import ResolvedStaticFile

from bfabric.entities import Dataset

if TYPE_CHECKING:
    from bfabric_app_runner.specs.inputs.bfabric_dataset_spec import BfabricDatasetSpec
    from bfabric import Bfabric


class ResolveBfabricDatasetSpecs:
    def __init__(self, client: Bfabric) -> None:
        self._client = client

    def __call__(self, specs: list[BfabricDatasetSpec]) -> list[ResolvedStaticFile]:
        """Convert dataset specifications to file specifications."""
        if not specs:
            return []

        # Fetch all datasets in bulk
        dataset_ids = [spec.id for spec in specs]
        datasets = Dataset.find_all(ids=dataset_ids, client=self._client)

        # Resolve each dataset specification to its serialized content
        return [
            ResolvedStaticFile(
                content=self._get_content(dataset=datasets[dataset_id], spec=spec), filename=spec.filename
            )
            for dataset_id, spec in zip(dataset_ids, specs)
        ]

    def _get_content(self, dataset: Dataset, spec: BfabricDatasetSpec) -> str | bytes:
        if spec.format == "csv":
            return dataset.get_csv(separator=spec.separator)
        elif spec.format == "parquet":
            return dataset.get_parquet()
        else:
            assert_never(spec.format)



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, assert_never

from bfabric_app_runner.inputs.resolve._resolve_bfabric_annotation_specs import ResolveBfabricAnnotationSpecs
from bfabric_app_runner.inputs.resolve._resolve_bfabric_dataset_specs import ResolveBfabricDatasetSpecs
from bfabric_app_runner.inputs.resolve._resolve_bfabric_order_fasta_specs import ResolveBfabricOrderFastaSpecs
from bfabric_app_runner.inputs.resolve._resolve_bfabric_resource_archive_specs import ResolveBfabricResourceArchiveSpecs
from bfabric_app_runner.inputs.resolve._resolve_bfabric_resource_specs import ResolveBfabricResourceSpecs
from bfabric_app_runner.inputs.resolve._resolve_file_specs import ResolveFileSpecs
from bfabric_app_runner.inputs.resolve._resolve_static_file_specs import ResolveStaticFileSpecs
from bfabric_app_runner.inputs.resolve._resolve_static_yaml_specs import ResolveStaticYamlSpecs
from bfabric_app_runner.inputs.resolve.resolved_inputs import ResolvedInputs
from bfabric_app_runner.specs.inputs.bfabric_annotation_spec import (
    BfabricAnnotationSpec,
)
from bfabric_app_runner.specs.inputs.bfabric_dataset_spec import BfabricDatasetSpec
from bfabric_app_runner.specs.inputs.bfabric_order_fasta_spec import BfabricOrderFastaSpec
from bfabric_app_runner.specs.inputs.bfabric_resource_archive_spec import BfabricResourceArchiveSpec
from bfabric_app_runner.specs.inputs.bfabric_resource_spec import BfabricResourceSpec
from bfabric_app_runner.specs.inputs.file_spec import FileSpec
from bfabric_app_runner.specs.inputs.static_file_spec import StaticFileSpec
from bfabric_app_runner.specs.inputs.static_yaml_spec import StaticYamlSpec

if TYPE_CHECKING:
    from bfabric import Bfabric
    from bfabric_app_runner.specs.inputs_spec import InputSpecType


class Resolver:
    """Resolves input specifications into standardized file specifications."""

    def __init__(self, client: Bfabric) -> None:
        self._client = client
        self._resolve_bfabric_dataset_specs = ResolveBfabricDatasetSpecs(client=client)
        self._resolve_bfabric_resource_specs = ResolveBfabricResourceSpecs(client=client)
        self._resolve_bfabric_resource_archive_specs = ResolveBfabricResourceArchiveSpecs(client=client)
        self._resolve_static_yaml_specs = ResolveStaticYamlSpecs()
        self._resolve_static_file_specs = ResolveStaticFileSpecs()
        self._resolve_bfabric_order_fasta_specs = ResolveBfabricOrderFastaSpecs(client=client)
        self._resolve_bfabric_annotation_specs = ResolveBfabricAnnotationSpecs(client=client)
        self._resolve_file_specs = ResolveFileSpecs()

    def resolve(self, specs: list[InputSpecType]) -> ResolvedInputs:
        """Convert input specifications to resolved file specifications."""
        grouped_specs = self._group_specs_by_type(specs=specs)
        files = []

        for spec_type, specs_list in grouped_specs.items():
            if issubclass(spec_type, StaticYamlSpec):
                files.extend(self._resolve_static_yaml_specs(specs_list))
            elif issubclass(spec_type, StaticFileSpec):
                files.extend(self._resolve_static_file_specs(specs_list))
            elif issubclass(spec_type, BfabricResourceSpec):
                files.extend(self._resolve_bfabric_resource_specs(specs_list))
            elif issubclass(spec_type, BfabricResourceArchiveSpec):
                files.extend(self._resolve_bfabric_resource_archive_specs(specs_list))
            elif issubclass(spec_type, BfabricDatasetSpec):
                files.extend(self._resolve_bfabric_dataset_specs(specs_list))
            elif issubclass(spec_type, BfabricOrderFastaSpec):
                files.extend(self._resolve_bfabric_order_fasta_specs(specs_list))
            elif issubclass(spec_type, BfabricAnnotationSpec):
                files.extend(self._resolve_bfabric_annotation_specs(specs_list))
            elif issubclass(spec_type, FileSpec):
                files.extend(self._resolve_file_specs(specs_list))
            else:
                assert_never(spec_type)

        return ResolvedInputs(files=files)

    @staticmethod
    def _group_specs_by_type(specs: list[InputSpecType]) -> dict[type, list]:
        """Group specifications by their type."""
        grouped = defaultdict(list)
        for spec in specs:
            grouped[type(spec)].append(spec)
        return grouped



from __future__ import annotations

from typing import Annotated, TYPE_CHECKING

import yaml
from bfabric_app_runner.specs.inputs.bfabric_annotation_spec import BfabricAnnotationSpec
from bfabric_app_runner.specs.inputs.bfabric_dataset_spec import BfabricDatasetSpec
from bfabric_app_runner.specs.inputs.bfabric_order_fasta_spec import BfabricOrderFastaSpec
from bfabric_app_runner.specs.inputs.bfabric_resource_archive_spec import BfabricResourceArchiveSpec
from bfabric_app_runner.specs.inputs.bfabric_resource_spec import BfabricResourceSpec
from bfabric_app_runner.specs.inputs.file_spec import FileSpec
from bfabric_app_runner.specs.inputs.static_file_spec import StaticFileSpec
from bfabric_app_runner.specs.inputs.static_yaml_spec import StaticYamlSpec
from pydantic import BaseModel, ConfigDict, Field

if TYPE_CHECKING:
    from pathlib import Path

InputSpecType = Annotated[
    BfabricResourceSpec
    | FileSpec
    | BfabricDatasetSpec
    | BfabricOrderFastaSpec
    | BfabricAnnotationSpec
    | BfabricResourceArchiveSpec
    | StaticYamlSpec
    | StaticFileSpec,
    Field(discriminator="type"),
]


class InputsSpec(BaseModel):
    model_config = ConfigDict(extra="forbid")
    inputs: list[InputSpecType]

    @classmethod
    def read_yaml(cls, path: Path) -> InputsSpec:
        return cls.model_validate(yaml.safe_load(path.read_text()))

    @classmethod
    def write_yaml(cls, inputs: list[InputSpecType], path: Path) -> None:
        model = cls(inputs=inputs)
        path.write_text(yaml.dump(model.model_dump(mode="json")))



from __future__ import annotations

from typing import TYPE_CHECKING

import yaml
from loguru import logger

from bfabric.entities import Resource
from bfabric.entities import Storage, Workunit
from bfabric.experimental.upload_dataset import bfabric_save_csv2dataset
from bfabric_app_runner.specs.outputs_spec import (
    CopyResourceSpec,
    UpdateExisting,
    OutputsSpec,
    SpecType,
    SaveDatasetSpec,
    SaveLinkSpec,
)
from bfabric_app_runner.util.checksums import md5sum
from bfabric_app_runner.util.scp import scp

if TYPE_CHECKING:
    from pathlib import Path
    from bfabric import Bfabric
    from bfabric.experimental.workunit_definition import WorkunitDefinition


def _get_output_folder(spec: CopyResourceSpec, workunit_definition: WorkunitDefinition) -> Path:
    if not spec.store_folder_path:
        return workunit_definition.registration.storage_output_folder
    else:
        return spec.store_folder_path


def register_file_in_workunit(
    spec: CopyResourceSpec,
    client: Bfabric,
    workunit_definition: WorkunitDefinition,
    resource_id: int | None = None,
) -> None:
    """Registers a file in the workunit."""
    existing_id = _identify_existing_resource_id(client, spec, workunit_definition)
    if resource_id is not None and existing_id is not None and resource_id != existing_id:
        raise ValueError(f"Resource id {resource_id} does not match existing resource id {existing_id}")

    checksum = md5sum(spec.local_path)
    output_folder = _get_output_folder(spec, workunit_definition=workunit_definition)
    resource_data = {
        "name": spec.store_entry_path.name,
        "workunitid": workunit_definition.registration.workunit_id,
        "storageid": workunit_definition.registration.storage_id,
        "relativepath": output_folder / spec.store_entry_path,
        "filechecksum": checksum,
        "status": "available",
        "size": spec.local_path.stat().st_size,
    }
    if resource_id is not None:
        resource_data["id"] = resource_id
    if existing_id is not None:
        resource_data["id"] = existing_id

    client.save("resource", resource_data)


def _identify_existing_resource_id(
    client: Bfabric, spec: CopyResourceSpec, workunit_definition: WorkunitDefinition
) -> int | None:
    """Returns the id of the existing resource if it exists."""
    if spec.update_existing in (UpdateExisting.IF_EXISTS, UpdateExisting.REQUIRED):
        # TODO maybe it would be more accurate to use relativepath here, however historically it would often start
        #      with `/` which can be confusing.
        resources = Resource.find_by(
            {
                "name": spec.store_entry_path.name,
                "workunitid": workunit_definition.registration.workunit_id,
            },
            client=client,
        ).values()
        if resources:
            return list(resources)[0].id
        elif spec.update_existing == UpdateExisting.REQUIRED:
            raise ValueError(f"Resource {spec.store_entry_path.name} not found in workunit {workunit_definition.id}")
    return None


def copy_file_to_storage(
    spec: CopyResourceSpec,
    workunit_definition: WorkunitDefinition,
    storage: Storage,
    ssh_user: str | None,
) -> None:
    """Copies a file to the storage, according to the spec."""
    # TODO here some direct uses of storage could still be optimized away
    output_folder = _get_output_folder(spec, workunit_definition=workunit_definition)
    output_uri = f"{storage.scp_prefix}{output_folder / spec.store_entry_path}"
    scp(spec.local_path, output_uri, user=ssh_user)


def _save_dataset(spec: SaveDatasetSpec, client: Bfabric, workunit_definition: WorkunitDefinition) -> None:
    """Saves a dataset to the bfabric."""
    # TODO should not print to stdout in the future
    # TODO also it should not be imported from bfabric_scripts, but rather the generic functionality should be available
    #      in the main package
    bfabric_save_csv2dataset(
        client=client,
        csv_file=spec.local_path,
        dataset_name=spec.name or spec.local_path.stem,
        container_id=workunit_definition.registration.container_id,
        workunit_id=workunit_definition.registration.workunit_id,
        sep=spec.separator,
        has_header=spec.has_header,
        invalid_characters=spec.invalid_characters,
    )


def _save_link(spec: SaveLinkSpec, client: Bfabric, workunit_definition: WorkunitDefinition) -> None:
    entity_id = spec.entity_id
    entity_type = spec.entity_type

    if entity_id is None:
        # double check, although it should be already checked by the validator
        if entity_type != "Workunit":
            raise ValueError("entity_id must be provided if entity_type is not 'Workunit'")
        entity_id = workunit_definition.registration.workunit_id

    # Check if the link already exists
    res = client.read("link", {"name": spec.name, "parentid": entity_id, "parentclassname": entity_type})
    existing_link_id = res[0]["id"] if len(res) > 0 else None
    # TODO maybe some of this logic could be extracted generically (i.e. UPDATE_EXISTING logic)
    if existing_link_id is not None and spec.update_existing == UpdateExisting.NO:
        msg = (
            f"Link {spec.name} already exists for entity {entity_type} with id {entity_id}, "
            f"but existing links should not be updated."
        )
        raise ValueError(msg)
    elif existing_link_id is None and spec.update_existing == UpdateExisting.REQUIRED:
        msg = (
            f"Link {spec.name} does not exist for entity {entity_type} with id {entity_id}, "
            f"but existing links is expected to be updated."
        )
        raise ValueError(msg)

    # Create or update the link
    link_data = {"name": spec.name, "url": spec.url, "parentid": entity_id, "parentclassname": entity_type}
    if existing_link_id is not None:
        link_data["id"] = existing_link_id
    res = client.save("link", link_data)
    logger.info(f"Link {spec.name} saved with id {res['id']} for entity {entity_type} with id {entity_id}")


def find_default_resource_id(workunit_definition: WorkunitDefinition, client: Bfabric) -> int | None:
    """Finds the default resource's id for the workunit. Maybe in the future, this will be always `None`."""
    workunit = Workunit.find(id=workunit_definition.registration.workunit_id, client=client)
    candidate_resources = [
        resource for resource in workunit.resources if resource["name"] not in ["slurm_stdout", "slurm_stderr"]
    ]
    # We also check that the resource is pending, as else we might re-use a resource that was created by the app...
    if len(candidate_resources) == 1 and candidate_resources[0]["status"] == "pending":
        return candidate_resources[0].id
    return None


def register_all(
    client: Bfabric,
    workunit_definition: WorkunitDefinition,
    specs_list: list[SpecType],
    ssh_user: str | None,
    reuse_default_resource: bool,
    force_storage: Path | None,
) -> None:
    """Registers all the output specs to the workunit."""
    default_resource_was_reused = not reuse_default_resource

    storage = _get_storage(client, force_storage, specs_list, workunit_definition)
    logger.info(f"Using storage: {storage}")

    for spec in specs_list:
        logger.debug(f"Registering {spec}")
        if isinstance(spec, CopyResourceSpec):
            copy_file_to_storage(
                spec,
                workunit_definition=workunit_definition,
                storage=storage,
                ssh_user=ssh_user,
            )
            if not default_resource_was_reused:
                resource_id = find_default_resource_id(workunit_definition=workunit_definition, client=client)
                default_resource_was_reused = True
            else:
                resource_id = None
            register_file_in_workunit(
                spec,
                client=client,
                workunit_definition=workunit_definition,
                resource_id=resource_id,
            )
        elif isinstance(spec, SaveDatasetSpec):
            _save_dataset(spec, client, workunit_definition=workunit_definition)
        elif isinstance(spec, SaveLinkSpec):
            _save_link(spec, client, workunit_definition=workunit_definition)
        else:
            raise ValueError(f"Unknown spec type: {type(spec)}")


def _get_storage(
    client: Bfabric, force_storage: Path | None, specs_list: list[SpecType], workunit_definition: WorkunitDefinition
) -> Storage | None:
    if any(isinstance(spec, CopyResourceSpec) for spec in specs_list):
        if force_storage is None:
            return Storage.find(workunit_definition.registration.storage_id, client=client)
        else:
            return Storage(yaml.safe_load(force_storage.read_text()), client=client)
            # TODO replace this later (to avoid versioning issues, i hardcode the above)
            # return Storage.load_yaml(force_storage, client=client)
    return None


def register_outputs(
    outputs_yaml: Path,
    workunit_definition: WorkunitDefinition,
    client: Bfabric,
    ssh_user: str | None,
    reuse_default_resource: bool,
    force_storage: Path | None,
) -> None:
    """Registers outputs to the workunit."""
    specs_list = OutputsSpec.read_yaml(outputs_yaml)
    register_all(
        client=client,
        workunit_definition=workunit_definition,
        specs_list=specs_list,
        ssh_user=ssh_user,
        reuse_default_resource=reuse_default_resource,
        force_storage=force_storage,
    )



import fnmatch
import shutil
import zipfile
from pathlib import Path

from bfabric_app_runner.inputs.resolve.resolved_inputs import ResolvedDirectory, ResolvedFile
from bfabric_app_runner.inputs.prepare.prepare_resolved_file import prepare_resolved_file
from loguru import logger


def prepare_resolved_directory(
    file: ResolvedDirectory,
    working_dir: Path,
    ssh_user: str | None,
) -> None:
    """Prepares the directory specified by the spec."""
    output_path = working_dir / file.filename
    output_path.parent.mkdir(exist_ok=True, parents=True)

    if file.extract == "zip":
        _prepare_zip_archive(file, output_path, ssh_user)
    else:
        raise NotImplementedError(f"Extraction type {file.extract} not supported")


def _prepare_zip_archive(file: ResolvedDirectory, output_path: Path, ssh_user: str | None) -> None:
    """Prepare a zip archive by downloading, extracting, and filtering."""
    # Download zip to permanent location in working directory for caching
    zip_filename = f"{file.filename}.zip"
    zip_path = output_path.parent / zip_filename
    _download_file(file, zip_path, ssh_user)
    _extract_zip_with_filtering(zip_path, output_path, file)


def _download_file(file: ResolvedDirectory, zip_path: Path, ssh_user: str | None) -> None:
    """Download the file from the specified source using existing file operations."""
    zip_resolved_file = ResolvedFile(
        source=file.source,
        filename=zip_path.name,
        link=False,
        checksum=None,
    )
    prepare_resolved_file(file=zip_resolved_file, working_dir=zip_path.parent, ssh_user=ssh_user)


def _extract_zip_with_filtering(zip_path: Path, output_path: Path, file: ResolvedDirectory) -> None:
    """Extract zip file with include/exclude filtering and optional root stripping."""
    output_path.mkdir(parents=True, exist_ok=True)

    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        # Filter files based on include/exclude patterns, excluding directories
        filtered_files = [
            f
            for f in _filter_files(zip_ref.namelist(), file.include_patterns, file.exclude_patterns)
            if not f.endswith("/")
        ]

        # Determine if we should strip root based on archive structure
        should_strip_root = file.strip_root and _should_strip_root_directory(zip_ref.namelist())

        # Extract filtered files
        for file_path in filtered_files:
            output_file_path = _get_output_file_path(file_path, output_path, should_strip_root)
            output_file_path.parent.mkdir(parents=True, exist_ok=True)

            logger.info(f"Extracting {file_path} to {output_file_path}")
            with zip_ref.open(file_path) as source, output_file_path.open("wb") as target:
                shutil.copyfileobj(source, target)


def _should_strip_root_directory(all_files: list[str]) -> bool:
    """Determine if we should strip the root directory based on archive structure."""
    if not all_files:
        return False

    # Get unique root-level entries (directories and files)
    root_entries = set()
    for file_path in all_files:
        if "/" in file_path:
            root_entries.add(file_path.split("/")[0])
        else:
            root_entries.add(file_path)

    # Only strip if there's exactly one root-level entry and it's a directory
    if len(root_entries) == 1:
        root_entry = next(iter(root_entries))
        # Check if this root entry is a directory (has files under it)
        return any(file_path.startswith(root_entry + "/") for file_path in all_files)

    return False


def _filter_files(files: list[str], include_patterns: list[str], exclude_patterns: list[str]) -> list[str]:
    """Filter files based on include and exclude patterns."""

    def should_include(file_path: str) -> bool:
        # If include patterns are specified, file must match at least one
        if include_patterns and not any(fnmatch.fnmatch(file_path, pattern) for pattern in include_patterns):
            return False
        # If exclude patterns are specified, file must not match any
        return not (exclude_patterns and any(fnmatch.fnmatch(file_path, pattern) for pattern in exclude_patterns))

    return [file_path for file_path in files if should_include(file_path)]


def _get_output_file_path(file_path: str, output_path: Path, strip_root: bool) -> Path:
    """Get the output file path, optionally stripping the root directory."""
    if strip_root:
        # Remove the first directory component if present
        path_parts = Path(file_path).parts
        relative_path = Path(*path_parts[1:]) if len(path_parts) > 1 else Path(file_path)
    else:
        relative_path = Path(file_path)

    return output_path / relative_path



from pathlib import Path

from rich.pretty import pprint

from bfabric import Bfabric
from bfabric.experimental.workunit_definition import WorkunitDefinition
from bfabric.utils.cli_integration import use_client
from bfabric_app_runner.output_registration.register import register_all, register_outputs
from bfabric_app_runner.specs.outputs_spec import CopyResourceSpec, UpdateExisting


def _get_workunit_definition(client: Bfabric, workunit_ref: int | Path) -> WorkunitDefinition:
    """Get the workunit with the given id and raises an error if it is not found."""
    workunit_ref = workunit_ref.resolve() if isinstance(workunit_ref, Path) else workunit_ref
    # TODO can we do better and provide a cache_file even here?
    return WorkunitDefinition.from_ref(workunit=workunit_ref, client=client, cache_file=None)


@use_client
def cmd_outputs_register(
    outputs_yaml: Path,
    workunit_ref: int | Path,
    *,
    ssh_user: str | None = None,
    force_storage: Path | None = None,
    client: Bfabric,
    # TODO
    reuse_default_resource: bool = True,
) -> None:
    """Register the output files of a workunit."""
    register_outputs(
        outputs_yaml=outputs_yaml,
        workunit_definition=_get_workunit_definition(client, workunit_ref),
        client=client,
        ssh_user=ssh_user,
        reuse_default_resource=reuse_default_resource,
        force_storage=force_storage,
    )


@use_client
def cmd_outputs_register_single_file(
    local_path: Path,
    *,
    workunit_ref: int | Path,
    store_entry_path: Path | None = None,
    store_folder_path: Path | None = None,
    update_existing: UpdateExisting = UpdateExisting.NO,
    ssh_user: str | None = None,
    force_storage: Path | None = None,
    reuse_default_resource: bool = False,
    client: Bfabric,
) -> None:
    """Register a single file in the workunit.

    In general, it is recommended to use the `register` command instead of this one and declare files using YAML.
    """
    if store_entry_path is None:
        store_entry_path = local_path.name

    spec = CopyResourceSpec(
        local_path=local_path,
        store_entry_path=store_entry_path,
        store_folder_path=store_folder_path,
        update_existing=update_existing,
    )
    pprint(spec)
    register_all(
        client=client,
        workunit_definition=_get_workunit_definition(client, workunit_ref),
        specs_list=[spec],
        ssh_user=ssh_user,
        reuse_default_resource=reuse_default_resource,
        force_storage=force_storage,
    )



from __future__ import annotations

from typing import Literal, TYPE_CHECKING, Self

from bfabric_app_runner.specs.common_types import RelativeFilePath, AbsoluteFilePath  # noqa: TC001
from pydantic import BaseModel, model_validator

if TYPE_CHECKING:
    from bfabric import Bfabric


class FileSourceLocal(BaseModel):
    local: AbsoluteFilePath

    def get_filename(self) -> str:
        return self.local.split("/")[-1]


class FileSourceSshValue(BaseModel):
    host: str
    path: AbsoluteFilePath


class FileSourceSsh(BaseModel):
    ssh: FileSourceSshValue

    def get_filename(self) -> str:
        return self.ssh.path.split("/")[-1]


class FileSpec(BaseModel):
    type: Literal["file"] = "file"
    source: FileSourceSsh | FileSourceLocal
    filename: RelativeFilePath | None = None  # TODO we cannot reuse the same type
    link: bool = False
    checksum: str | None = None

    @model_validator(mode="after")
    def validate_no_link_ssh(self) -> Self:
        if isinstance(self.source, FileSourceSsh) and self.link:
            raise ValueError("Cannot link to a remote file.")
        return self

    def get_filename(self) -> str:
        """Returns the filename, extracting it from the source if it was omitted."""
        return self.filename if self.filename else self.source.get_filename()

    def resolve_filename(self, client: Bfabric) -> str:
        # TODO delete
        return self.get_filename()



from __future__ import annotations

from typing import Literal

from pydantic import BaseModel, ConfigDict

from bfabric_app_runner.specs.common_types import RelativeFilePath  # noqa: TC001


class BfabricDatasetSpec(BaseModel):
    model_config = ConfigDict(extra="forbid")
    type: Literal["bfabric_dataset"] = "bfabric_dataset"

    id: int
    """B-Fabric dataset ID"""

    filename: RelativeFilePath
    """Target filename to save to"""

    separator: Literal[",", "\t"] = ","
    """Separator for the CSV file (not relevant for Parquet)"""

    format: Literal["csv", "parquet"] = "csv"



from bfabric_app_runner.inputs.resolve.resolved_inputs import ResolvedFile
from bfabric_app_runner.specs.inputs.file_spec import FileSpec


class ResolveFileSpecs:
    def __call__(self, specs: list[FileSpec]) -> list[ResolvedFile]:
        """Converts file specifications to resolved file specifications."""
        return [
            ResolvedFile(source=spec.source, filename=spec.get_filename(), link=spec.link, checksum=spec.checksum)
            for spec in specs
        ]



from __future__ import annotations

from typing import Literal, TYPE_CHECKING, Annotated

from bfabric_app_runner.specs.common_types import RelativeFilePath  # noqa: TC001
from pydantic import BaseModel, Field

if TYPE_CHECKING:
    from bfabric import Bfabric


class _AnnotationSpec(BaseModel):
    type: Literal["bfabric_annotation"] = "bfabric_annotation"
    filename: RelativeFilePath

    def resolve_filename(self, client: Bfabric) -> str:
        return self.filename


class BfabricAnnotationResourceSampleSpec(_AnnotationSpec):
    annotation: Literal["resource_sample"] = "resource_sample"
    separator: str
    resource_ids: list[int]
    format: Literal["csv"] = "csv"


BfabricAnnotationSpec = BfabricAnnotationResourceSampleSpec
BfabricAnnotationSpecField = Annotated[BfabricAnnotationSpec, Field(discriminator="annotation")]



import os
import shlex
import subprocess
from pathlib import Path

from loguru import logger

from bfabric_app_runner.specs.app.commands_spec import CommandExec


def _get_shell_env(
    environ: dict[str, str] | None, config_env: dict[str, str], config_prepend_paths: list[Path]
) -> dict[str, str]:
    if environ is None:
        environ = os.environ.copy()

    if config_prepend_paths:
        # Ensure environ['PATH'] is set
        environ["PATH"] = environ.get("PATH", "")
        # Prepend paths
        for path in reversed(config_prepend_paths):
            resolved_path = path.expanduser().absolute()
            environ["PATH"] = f"{resolved_path}:{environ['PATH']}"

    for key, value in config_env.items():
        environ[key] = value

    return environ


def execute_command_exec(command: CommandExec, *args: str, environ: dict[str, str] | None = None) -> None:
    """Executes the command with the provided arguments."""
    command_args = shlex.split(command.command) + list(args)
    shell_env = _get_shell_env(environ, command.env, command.prepend_paths)
    logger.info(f"Executing command: {shlex.join(command_args)}")
    logger.debug(f"{command_args=}")
    logger.trace(f"{shell_env=}")
    subprocess.run(command_args, check=True, env=shell_env)



from typing import Annotated

import cyclopts
from cyclopts import Parameter

from bfabric import Bfabric
from bfabric.utils.cli_integration import use_client
from bfabric_app_runner.actions.execute import (
    execute_dispatch,
    execute_run,
    execute_inputs,
    execute_process,
    execute_outputs,
)
from bfabric_app_runner.actions.types import ActionInputs, ActionProcess, ActionOutputs, ActionRun, ActionDispatch

app = cyclopts.App()


@use_client
def cmd_action_dispatch(*, action: Annotated[ActionDispatch, Parameter(name="*")], client: Bfabric) -> None:
    """Dispatches a workunit definition."""
    execute_dispatch(action, client=client)


@use_client
def cmd_action_run_all(*, action: Annotated[ActionRun, Parameter(name="*")], client: Bfabric) -> None:
    """Runs all stages of an app that has been dispatched."""
    execute_run(action, client=client)


@use_client
def cmd_action_inputs(*, action: Annotated[ActionInputs, Parameter(name="*")], client: Bfabric) -> None:
    """Prepares the input files by downloading and generating them (if necessary)."""
    execute_inputs(action, client=client)


@use_client
def cmd_action_process(*, action: Annotated[ActionProcess, Parameter(name="*")], client: Bfabric) -> None:
    """Processes a chunk."""
    execute_process(action, client=client)


@use_client
def cmd_action_outputs(*, action: Annotated[ActionOutputs, Parameter(name="*")], client: Bfabric) -> None:
    """Registers the output files of a chunk."""
    execute_outputs(action, client=client)



from __future__ import annotations

from typing import assert_never, TYPE_CHECKING

from bfabric_app_runner.inputs.resolve.resolved_inputs import ResolvedStaticFile
from loguru import logger

from bfabric.entities import Workunit, Order

if TYPE_CHECKING:
    from bfabric import Bfabric
    from bfabric_app_runner.specs.inputs.bfabric_order_fasta_spec import BfabricOrderFastaSpec


class ResolveBfabricOrderFastaSpecs:
    def __init__(self, client: Bfabric) -> None:
        self._client = client

    def __call__(self, specs: list[BfabricOrderFastaSpec]) -> list[ResolvedStaticFile]:
        """Convert order FASTA specifications to file specifications."""
        # Note: This approach is not efficient if there are multiple entries, but usually we only have one so it is
        #       not optimized yet.
        return [ResolvedStaticFile(content=self._get_order_fasta(spec=spec), filename=spec.filename) for spec in specs]

    def _get_order_fasta(self, spec: BfabricOrderFastaSpec) -> str:
        """Extract FASTA sequence from an order or workunit."""
        if spec.entity == "workunit":
            workunit = Workunit.find(id=spec.id, client=self._client)
            if not isinstance(workunit.container, Order):
                msg = f"Workunit {spec.id} is not associated with an order"
                if spec.required:
                    raise ValueError(msg)
                logger.warning(msg)
                return ""
            order = workunit.container
        elif spec.entity == "order":
            order = Order.find(id=spec.id, client=self._client)
        else:
            assert_never(spec.entity)

        return order.data_dict.get("fastasequence", "")



from pathlib import Path

from loguru import logger
from pydantic import BaseModel, ConfigDict

from bfabric import Bfabric
from bfabric.entities import Resource
from bfabric.experimental.workunit_definition import WorkunitDefinition
from bfabric_app_runner.dispatch.generic import (
    write_chunks_file,
)
from bfabric_app_runner.dispatch.resource_flow import get_resource_flow_input_resources


class ConfigDispatchSingleResourceFlow(BaseModel):
    model_config = ConfigDict(extra="forbid")
    filter_suffix: str | None = None


class DispatchSingleResourceFlow:
    def __init__(self, client: Bfabric, config: ConfigDispatchSingleResourceFlow, out_dir: Path) -> None:
        self._client = client
        self._config = config
        self._out_dir = out_dir

    def dispatch_job(self, resources: list[Resource], workunit: WorkunitDefinition) -> Path:
        raise NotImplementedError

    def dispatch_workunit(self, definition: WorkunitDefinition) -> None:
        if not definition.execution.resources:
            logger.error("No resources found in workunit.")
            return
        resources = get_resource_flow_input_resources(
            client=self._client,
            definition=definition,
            filter_suffix=self._config.filter_suffix,
        )
        path = self.dispatch_job(resources=resources, workunit=definition)
        write_chunks_file(out_dir=self._out_dir, chunks=[path])



import copy
from pathlib import Path

import yaml

from bfabric import Bfabric
from bfabric.experimental.workunit_definition import WorkunitDefinition
from bfabric.utils.cli_integration import use_client
from bfabric_app_runner.actions.config_file import ActionConfig
from bfabric_app_runner.prepare.makefile_template import render_makefile
from bfabric_app_runner.specs.app.app_spec import AppSpec


def _update_app_version(workunit_definition: WorkunitDefinition, application_version: str) -> WorkunitDefinition:
    # TODO if this is useful consider moving it to the WorkunitDefinition class
    workunit_definition = copy.deepcopy(workunit_definition)
    workunit_definition.execution.raw_parameters["application_version"] = application_version
    return workunit_definition


@use_client
def cmd_prepare_workunit(
    app_spec: Path,
    work_dir: Path,
    workunit_ref: int | Path,
    *,
    ssh_user: str | None = None,
    force_storage: Path | None = None,
    force_app_version: str | None = None,
    read_only: bool = False,
    use_external_runner: bool = False,
    client: Bfabric,
) -> None:
    """Prepares a workunit for processing.

    :param app_spec: The path to the application specification.
    :param work_dir: The directory where the workunit execution will be executed later.
    :param workunit_ref: The reference to the workunit, can be an ID or a path to a YAML file.
    :param ssh_user: The SSH user to use for file operations (if not provided, defaults to the current user).
    :param force_storage: (for testing) storage definition to use for the workunit output instead of the configured one.
    :param force_app_version: (for testing) specific application version instead of the workunit's `application_version`
    :param read_only: If True, no results will be written to B-Fabric.
    :param use_external_runner: If True, allows using external bfabric-app-runner from PATH instead of exact version.
    """
    work_dir.mkdir(parents=True, exist_ok=True)

    workunit_definition = WorkunitDefinition.from_ref(workunit_ref, client=client)
    if force_app_version:
        workunit_definition = _update_app_version(
            workunit_definition=workunit_definition, application_version=force_app_version
        )

    # Resolve the app version early, following the pattern used by other commands
    app_full_spec = AppSpec.load_yaml(
        app_yaml=app_spec,
        app_id=workunit_definition.registration.application_id,
        app_name=workunit_definition.registration.application_name,
    )

    workunit_definition_path = work_dir / "workunit_definition.yml"
    workunit_definition.to_yaml(path=workunit_definition_path)
    _write_app_env_file(
        path=work_dir / "app_env.yml",
        app_ref=app_spec.resolve(),
        workunit_ref=workunit_definition_path,
        ssh_user=ssh_user,
        force_storage=force_storage,
        read_only=read_only,
    )
    # Render the workunit Makefile template
    render_makefile(
        path=work_dir / "Makefile",
        bfabric_app_spec=app_full_spec.bfabric,
        rename_existing=True,
        use_external_runner=use_external_runner,
    )


def _write_app_env_file(
    path: Path,
    app_ref: Path,
    workunit_ref: Path,
    ssh_user: str | None,
    force_storage: Path | None,
    read_only: bool,
) -> None:
    # TODO this handles everything except the BFABRIPY_CONFIG_OVERRIDE

    action_config = ActionConfig(
        work_dir=path.parent.resolve(),
        app_ref=app_ref,
        workunit_ref=workunit_ref,
        ssh_user=ssh_user,
        force_storage=force_storage,
        read_only=read_only,
    )

    data = {"bfabric_app_runner": {"action": action_config.model_dump(mode="json")}}
    with path.open("w") as fh:
        yaml.safe_dump(data, fh)



from __future__ import annotations

import shlex
import subprocess
from pathlib import Path

from loguru import logger


def scp(source: str | Path, target: str | Path, *, user: str | None = None, mkdir: bool = True) -> None:
    """Copies a file using scp from source to target.
    Make sure that either the source or target specifies a host, otherwise you should just use shutil.copyfile.
    Always specify the full path for the target, not just a directory.
    """
    source = str(source)
    target = str(target)
    source_remote = _is_remote(source)
    target_remote = _is_remote(target)
    if source_remote == target_remote:
        msg = (
            f"Either source or target should be remote, but not both. "
            f"source_remote={repr(source_remote)} == target_remote={repr(target_remote)}"
        )
        raise ValueError(msg)
    if target[-1] == "/":
        raise ValueError(f"Target should be a file, not a directory: {target}")
    if user and source_remote:
        source = f"{user}@{source}"
    elif user and target_remote:
        target = f"{user}@{target}"

    if mkdir:
        if target_remote:
            host, path = target.split(":", 1)
            parent_path = Path(path).parent
            logger.debug(f"ssh {host} mkdir -p {parent_path}")
            subprocess.run(["ssh", host, "mkdir", "-p", parent_path], check=True)
        else:
            parent_path = Path(target).parent
            parent_path.mkdir(parents=True, exist_ok=True)

    cmd = ["scp", source, target]
    logger.info(shlex.join(cmd))
    subprocess.run(cmd, check=True)


def _is_remote(path: str | Path) -> bool:
    return ":" in str(path)



from __future__ import annotations
from pathlib import Path
from typing import TYPE_CHECKING

import mako.template
from loguru import logger
from pydantic import BaseModel


if TYPE_CHECKING:
    from bfabric.entities import Workunit


class WrapAppYamlTemplate:
    class Params(BaseModel):
        workunit_id: int
        app_yaml_path: str
        scratch_root: Path

        @classmethod
        def extract_workunit(cls, workunit: Workunit, scratch_root: Path) -> WrapAppYamlTemplate.Params:
            app_yaml_path = workunit.application.executable["program"]
            return cls(
                workunit_id=workunit.id,
                app_yaml_path=app_yaml_path,
                scratch_root=scratch_root,
            )

    def __init__(self, params: Params, path: Path) -> None:
        self._params = params
        self._path = path

    @classmethod
    def default_path(cls) -> Path:
        return Path(__file__).parent / "wrap_app_yaml_template.bash.mako"

    def render_string(self) -> str:
        params = self._params.model_dump(mode="python")
        logger.debug("Rendering {} with params: {}", self._path, params)
        template = mako.template.Template(filename=str(self._path))
        return template.render(**params)



from pathlib import Path
from typing import Any

import yaml
from glom import glom
from loguru import logger

from pydantic import BaseModel, model_validator


class ActionConfig(BaseModel):
    """Configuration which will be passed to the corresponding action."""

    work_dir: Path | None = None
    app_ref: Path | None = None
    workunit_ref: int | Path | None = None
    ssh_user: str | None = None
    filter: str | None = None
    force_storage: Path | None = None
    read_only: bool | None = None


class FromConfigFile(BaseModel):
    config: Path | None = None

    @model_validator(mode="before")
    @classmethod
    def parse_config_file(cls, values: Any) -> Any:
        if isinstance(values, dict) and "config" in values and values["config"] is not None:
            logger.debug(f"Parsing config file: {values['config']}")
            with Path(values["config"]).open("r") as config_file:
                config_data = yaml.safe_load(config_file)
            config_entry = glom(config_data, "bfabric_app_runner.action")
            data = ActionConfig.model_validate(config_entry)

            # update all values which are not provided explicitly with the values from config_parsed
            for key, value in data:
                if key not in values and value is not None:
                    logger.debug(f"Overriding {key} with value from config file: {value}")
                    values[key] = value

        return values



import subprocess
from pathlib import Path

from bfabric import Bfabric
from bfabric.experimental.workunit_definition import WorkunitDefinition
from bfabric.utils.cli_integration import use_client
from bfabric_app_runner.cli.cmd_prepare import cmd_prepare_workunit


@use_client
def cmd_run_workunit(app_definition: Path, scratch_root: Path, workunit_ref: int | Path, *, client: Bfabric) -> None:
    """Runs a workunit by preparing the workunit and executing it."""
    workunit_definition = WorkunitDefinition.from_ref(workunit_ref, client=client)
    try:
        client.save("workunit", {"id": workunit_definition.registration.workunit_id, "status": "processing"})
        work_dir = (
            scratch_root
            / f"A{workunit_definition.registration.application_id}_{workunit_definition.registration.application_name}"
            / f"WU{workunit_definition.registration.workunit_id}"
        )
        work_dir.mkdir(parents=True, exist_ok=True)
        # TODO caching of workunit_definition
        cmd_prepare_workunit(app_spec=app_definition, work_dir=work_dir, workunit_ref=workunit_ref)
        # TODO directly call through python
        subprocess.run(["make", "run-all"], cwd=work_dir, check=True)
        client.save("workunit", {"id": workunit_definition.registration.workunit_id, "status": "available"})
    except Exception:
        client.save("workunit", {"id": workunit_definition.registration.workunit_id, "status": "failed"})
        raise



from __future__ import annotations

from functools import cached_property
from pathlib import Path  # noqa: TC003

import yaml
from pydantic import BaseModel, field_validator

from bfabric.entities import Workunit  # noqa: TC002
from bfabric.utils.path_safe_name import path_safe_name
from bfabric_app_runner.bfabric_integration.submitter.config.slurm_workunit_params import (
    SlurmWorkunitParams,  # noqa: TC001
)  # noqa: TC001
from bfabric_app_runner.specs.config_interpolation import VariablesApp, VariablesWorkunit, interpolate_config_strings


class SlurmParameters(BaseModel):
    """The concrete Slurm parameters for a specific workunit."""

    # Note: Originally, the idea was that at the app level app_params could also be defined, but this was omitted for
    #       simplicity in this version.
    submitter_params: dict[str, str | int | None]
    """Allows setting arbitrary parameters."""
    job_script: Path
    """The path to store job script."""
    workunit_params: SlurmWorkunitParams
    """Allows setting a controlled set of parameters."""
    scratch_root: Path
    """The root directory for scratch space."""

    @cached_property
    def sbatch_params(self) -> dict[str, str]:
        merged = {**self.submitter_params, **self.workunit_params.as_dict()}
        return {key: str(value) for key, value in merged.items() if value is not None}


class _SlurmConfigFile(BaseModel):
    params: dict[str, str | int | None]
    job_script: Path
    scratch_root: Path

    @field_validator("job_script", "scratch_root", mode="after")
    @classmethod
    def expand_user_in_paths(cls, value: Path) -> Path:
        """Expands user in paths."""
        return value.expanduser()


class _SlurmConfigFileTemplate(BaseModel):
    """The generic slurm configuration file, with template strings not yet evaluated."""

    params: dict[str, str | int | None]
    job_script: str
    scratch_root: Path

    @classmethod
    def for_yaml(cls, path: Path) -> _SlurmConfigFileTemplate:
        return _SlurmConfigFileTemplate.model_validate(yaml.safe_load(path.read_text()))

    def evaluate(self, app: VariablesApp, workunit: VariablesWorkunit) -> _SlurmConfigFile:
        data_template = self.model_dump(mode="json")
        data = interpolate_config_strings(data=data_template, variables={"app": app, "workunit": workunit})
        return _SlurmConfigFile.model_validate(data)


def evaluate_slurm_parameters(config_yaml_path: Path, workunit: Workunit) -> SlurmParameters:
    """Evaluates the Slurm Parameters from the YAML file for a given workunit."""
    config_file_template = _SlurmConfigFileTemplate.for_yaml(config_yaml_path)
    app_variables = VariablesApp(
        id=workunit.application.id, name=path_safe_name(workunit.application["name"]), version="latest"
    )
    workunit_variables = VariablesWorkunit(id=workunit.id)
    config_file = config_file_template.evaluate(app=app_variables, workunit=workunit_variables)
    workunit_params = SlurmWorkunitParams.model_validate(workunit.submitter_parameters)
    return SlurmParameters(
        submitter_params=config_file.params,
        job_script=config_file.job_script,
        workunit_params=workunit_params,
        scratch_root=config_file.scratch_root,
    )



from __future__ import annotations

from typing import TYPE_CHECKING

import yaml
from bfabric_app_runner.inputs.resolve.resolved_inputs import ResolvedStaticFile

if TYPE_CHECKING:
    from bfabric_app_runner.specs.inputs.static_yaml_spec import StaticYamlSpec


class ResolveStaticYamlSpecs:
    def __call__(self, specs: list[StaticYamlSpec]) -> list[ResolvedStaticFile]:
        """Convert YAML specifications to file specifications."""
        return [ResolvedStaticFile(content=yaml.safe_dump(spec.data), filename=spec.filename) for spec in specs]



from __future__ import annotations

from typing import Literal

from bfabric_app_runner.specs.common_types import RelativeFilePath
from pydantic import BaseModel, ConfigDict


class BfabricResourceArchiveSpec(BaseModel):
    model_config = ConfigDict(extra="forbid")
    type: Literal["bfabric_resource_archive"] = "bfabric_resource_archive"

    id: int
    """B-Fabric resource ID"""

    filename: RelativeFilePath
    """Target directory to save to"""

    extract: Literal["zip"] = "zip"
    """Extraction to perform, currently only 'zip' is supported"""

    include_patterns: list[str] = []
    """Globs of files to include in the archive extraction (by default all files are included)"""

    exclude_patterns: list[str] = []
    """Globs of files to exclude from the archive extraction (by default no files are excluded)"""

    strip_root: bool = False
    """If True, the root directory (if present) of the archive will be stripped during extraction."""

    check_checksum: bool = True
    """Whether to check the checksum of the archive file, after downloading."""



from typing import Literal

from pydantic import BaseModel


class StaticYamlSpec(BaseModel):
    type: Literal["static_yaml"] = "static_yaml"

    data: dict | list
    """The YAML document content to write."""
    filename: str
    """The target filename to write to."""



from __future__ import annotations

from typing import TYPE_CHECKING

from bfabric_app_runner.inputs.resolve.resolved_inputs import ResolvedStaticFile

if TYPE_CHECKING:
    from bfabric_app_runner.specs.inputs.static_file_spec import StaticFileSpec


class ResolveStaticFileSpecs:
    def __call__(self, specs: list[StaticFileSpec]) -> list[ResolvedStaticFile]:
        """Convert YAML specifications to file specifications."""
        return [ResolvedStaticFile(content=spec.content, filename=spec.filename) for spec in specs]



from pathlib import Path

from bfabric import Bfabric
from bfabric.utils.cli_integration import use_client
from bfabric_app_runner.inputs.prepare.prepare_folder import prepare_folder
from bfabric_app_runner.inputs.list_inputs.list_inputs import (
    list_input_states,
    print_input_states,
    FileState,
    IntegrityState,
)
from bfabric_app_runner.specs.inputs_spec import InputsSpec


@use_client
def cmd_inputs_prepare(
    inputs_yaml: Path,
    target_folder: Path | None = None,
    *,
    ssh_user: str | None = None,
    filter: str | None = None,
    client: Bfabric,
) -> None:
    """Prepare the input files by downloading and generating them (if necessary).

    :param inputs_yaml: Path to the inputs.yml file.
    :param target_folder: Path to the target folder where the input files should be downloaded.
    :param ssh_user: SSH user to use for downloading the input files, instead of the current user.
    :param filter: only this input file will be prepared.
    """
    prepare_folder(
        inputs_yaml=inputs_yaml,
        target_folder=target_folder,
        ssh_user=ssh_user,
        client=client,
        action="prepare",
        filter=filter,
    )


@use_client
def cmd_inputs_clean(
    inputs_yaml: Path,
    target_folder: Path | None = None,
    *,
    filter: str | None = None,
    client: Bfabric,
) -> None:
    """Removes all local copies of input files.

    :param inputs_yaml: Path to the inputs.yml file.
    :param target_folder: Path to the target folder where the input files should be removed.
    :param filter: only this input file will be removed.
    """
    # TODO clean shouldn't even need all these arguments, this could be refactored later
    prepare_folder(
        inputs_yaml=inputs_yaml,
        target_folder=target_folder,
        ssh_user=None,
        client=client,
        action="clean",
        filter=filter,
    )


def get_inputs_and_print(
    inputs_yaml: Path,
    target_folder: Path | None,
    check: bool,
    client: Bfabric,
) -> list[FileState]:
    """Reads the input files, performing integrity checks if requested, and prints the results."""
    input_states = list_input_states(
        specs=InputsSpec.read_yaml(inputs_yaml).inputs,
        target_folder=target_folder or Path(),
        client=client,
        check_files=check,
    )
    print_input_states(input_states)
    return input_states


@use_client
def cmd_inputs_list(
    inputs_yaml: Path,
    target_folder: Path | None = None,
    check: bool = False,
    *,
    client: Bfabric,
) -> None:
    """Lists the input files for an app.

    :param inputs_yaml: Path to the inputs.yml file.
    :param target_folder: Path to the target folder where the input files should be located, if different from the
        file containing the inputs.yml file.
    """
    get_inputs_and_print(inputs_yaml=inputs_yaml, target_folder=target_folder, check=check, client=client)


@use_client
def cmd_inputs_check(
    inputs_yaml: Path,
    target_folder: Path | None = None,
    *,
    client: Bfabric,
) -> None:
    """Checks if the input files are present and have the correct content.

    The script will exit with a non-zero status
    code if any of the input files are missing or have incorrect content.
    :param inputs_yaml: Path to the inputs.yml file.
    :param target_folder: Path to the target folder where the input files should be located, if different from the
        file containing the inputs.yml file.
    """
    input_states = get_inputs_and_print(inputs_yaml=inputs_yaml, target_folder=target_folder, check=True, client=client)
    invalid_states = {state.integrity for state in input_states if state.integrity != IntegrityState.Correct}
    if invalid_states:
        print(f"Encountered invalid input states: {invalid_states}")
        raise SystemExit(1)
    else:
        print("All input files are correct.")



from __future__ import annotations

import io
from typing import TYPE_CHECKING, assert_never

import polars as pl
from bfabric_app_runner.inputs.resolve.resolved_inputs import ResolvedStaticFile

from bfabric.entities import Resource
from bfabric.utils.polars_utils import flatten_relations

if TYPE_CHECKING:
    from bfabric import Bfabric
    from bfabric_app_runner.specs.inputs.bfabric_annotation_spec import (
        BfabricAnnotationSpec,
        BfabricAnnotationResourceSampleSpec,
    )


class ResolveBfabricAnnotationSpecs:
    def __init__(self, client: Bfabric) -> None:
        self._client = client

    def __call__(self, specs: list[BfabricAnnotationSpec]) -> list[ResolvedStaticFile]:
        """Convert annotation specifications to file specifications."""
        # Note: This approach is not efficient if there are multiple entries, but usually we only have one so it is
        #       not optimized yet.
        return [
            ResolvedStaticFile(content=get_annotation(spec=spec, client=self._client), filename=spec.filename)
            for spec in specs
        ]


def _get_resource_sample_annotation(spec: BfabricAnnotationResourceSampleSpec, client: Bfabric) -> pl.DataFrame:
    """Returns the annotation content for the resource_sample annotation type."""
    # load entities
    resources = list(Resource.find_all(spec.resource_ids, client).values())
    samples = [resource.sample for resource in resources]

    # flatten and merge data
    resources_df = flatten_relations(
        pl.DataFrame([resource.data_dict for resource in resources]).select(pl.all().name.prefix("resource_"))
    )
    samples_df = flatten_relations(
        pl.DataFrame([sample.data_dict for sample in samples]).select(pl.all().name.prefix("sample_"))
    )
    return resources_df.join(samples_df, left_on="resource_sample_id", right_on="sample_id", how="left")


def get_annotation(spec: BfabricAnnotationSpec, client: Bfabric) -> str | bytes:
    """Returns the annotation content specified by the spec."""
    match spec.annotation:
        case "resource_sample":
            annotation_df = _get_resource_sample_annotation(spec, client=client)
        case _:
            raise ValueError(f"Unsupported annotation type: {spec.annotation}")
    if spec.format == "csv":
        return annotation_df.write_csv(separator=spec.separator)
    elif spec.format == "parquet":
        bytes_io = io.BytesIO()
        annotation_df.write_parquet(bytes_io)
        return bytes_io.getvalue()
    else:
        assert_never(spec.format)



import shlex
import shutil
import subprocess
from pathlib import Path
from shutil import SameFileError
from subprocess import CalledProcessError
from typing import assert_never

from bfabric_app_runner.inputs.resolve.resolved_inputs import ResolvedFile
from bfabric_app_runner.specs.inputs.file_spec import (
    FileSourceSsh,
    FileSourceLocal,
    FileSourceSshValue,
)
from bfabric_app_runner.util.scp import scp
from loguru import logger


def prepare_resolved_file(file: ResolvedFile, working_dir: Path, ssh_user: str | None) -> None:
    """Prepares the file specified by the spec."""
    output_path = working_dir / file.filename
    output_path.parent.mkdir(exist_ok=True, parents=True)

    if not file.link:
        success = _operation_copy_rsync(file, output_path, ssh_user)
        if not success:
            success = _operation_copy(file, output_path, ssh_user)
    else:
        success = _operation_link_symbolic(file, output_path)
    if not success:
        raise RuntimeError(f"Failed to copy file: {file}")


def _operation_copy_rsync(file: ResolvedFile, output_path: Path, ssh_user: str | None) -> bool:
    match file.source:
        case FileSourceLocal(local=local):
            source_str = str(Path(local).resolve())
        case FileSourceSsh(ssh=FileSourceSshValue(host=host, path=path)):
            source_str = f"{ssh_user}@{host}:{path}" if ssh_user else f"{host}:{path}"
        case _:
            assert_never(file.source)
    cmd = ["rsync", "-Pav", source_str, str(output_path)]
    logger.info(shlex.join(cmd))
    result = subprocess.run(cmd, check=False)
    return result.returncode == 0


def _operation_copy(file: ResolvedFile, output_path: Path, ssh_user: str | None) -> bool:
    match file.source:
        case FileSourceLocal():
            return _operation_copy_cp(file, output_path)
        case FileSourceSsh():
            return _operation_copy_scp(file, output_path, ssh_user)
        case _:
            assert_never(file.source)


def _operation_copy_scp(file: ResolvedFile, output_path: Path, ssh_user: str | None) -> bool:
    try:
        source_str = f"{file.source.ssh.host}:{file.source.ssh.path}"
        scp(source=source_str, target=output_path, user=ssh_user)
    except CalledProcessError:
        return False
    return True


def _operation_copy_cp(file: ResolvedFile, output_path: Path) -> bool:
    cmd = [str(Path(file.source.local).resolve()), str(output_path)]
    logger.info(shlex.join(["cp", *cmd]))
    try:
        shutil.copyfile(*cmd)
    except (OSError, SameFileError):
        return False
    return True


def _operation_link_symbolic(file: ResolvedFile, output_path: Path) -> bool:
    # the link is created relative to the output file, so it should be more portable across apptainer images etc
    source_path = Path(file.source.local).resolve().relative_to(output_path.resolve().parent, walk_up=True)

    # if the file exists, and only if it is a link as well
    if output_path.is_symlink():
        # check if it points to the same file, in which case we don't need to do anything
        if output_path.resolve() == source_path.resolve():
            logger.info("Link already exists and points to the correct file")
            return True
        else:
            logger.info(f"rm {output_path}")
            output_path.unlink()
    elif output_path.exists():
        raise RuntimeError(f"Output path already exists and is not a symlink: {output_path}")
    cmd = ["ln", "-s", str(source_path), str(output_path)]
    logger.info(shlex.join(cmd))
    result = subprocess.run(cmd, check=False)
    return result.returncode == 0



from loguru import logger

from bfabric import Bfabric
from bfabric.entities import Resource
from bfabric.experimental.workunit_definition import WorkunitDefinition


def get_resource_flow_input_resources(
    client: Bfabric,
    definition: WorkunitDefinition,
    filter_suffix: str | None,
) -> list[Resource]:
    """Returns the input resources for a resource flow workunit, applying e.g. a filter suffix."""
    all_resources = Resource.find_all(ids=definition.execution.resources, client=client)
    result_resources = []
    for resource in sorted(all_resources.values()):
        if filter_suffix is not None and not resource["relativepath"].endswith(filter_suffix):
            logger.info(f"Skipping resource {resource['relativepath']!r} as it does not match the extension filter.")
            continue
        result_resources.append(resource)
    return result_resources



from __future__ import annotations

from typing import TYPE_CHECKING
from collections import Counter

import yaml
from pydantic import BaseModel, field_validator

from bfabric_app_runner.specs.app.app_version import AppVersion, AppVersionMultiTemplate  # noqa: TCH001
from bfabric_app_runner.specs.config_interpolation import VariablesApp

if TYPE_CHECKING:
    from pathlib import Path


class BfabricAppSpec(BaseModel):
    """Contains the app specification information that is relevant to bfabric, and not exactly the app itself."""

    app_runner: str
    """Specifies the app runner version to use for the app.

    We support both a PyPI version (e.g. `0.0.17`) as well as a git reference, which is a string in the
    format `git+https://github.com/fgcz/bfabricPy@main#subdirectory=bfabric_app_runner` where you can specify
    any git reference instead of `main` as needed.
    """

    workflow_template_step_id: int | None = None
    """If specified, this indicates that a workflow step should be created for the workunit."""


class AppSpecTemplate(BaseModel):
    """This model defines the app_spec definition in a file.

    As the name suggests, this is a template that can be expanded to a concrete ``AppSpec`` instance.
    The main difference is that this may contain usages of `Variables`. TODO
    """

    bfabric: BfabricAppSpec
    versions: list[AppVersionMultiTemplate]

    # TODO this should take the variables as param instead
    def evaluate(self, app_id: int, app_name: str) -> AppSpec:
        """Evaluates the template to a concrete ``AppSpec`` instance."""
        version_templates = [expanded for version in self.versions for expanded in version.expand_versions()]
        versions = []
        for version_template in version_templates:
            variables_app = VariablesApp(id=app_id, name=app_name, version=version_template.version)
            versions.append(version_template.evaluate(variables_app=variables_app))
        # TODO add interpolation for bfabric config for consistency
        return AppSpec.model_validate({"versions": versions, "bfabric": self.bfabric})


class AppSpec(BaseModel):
    """Parsed app versions from the app spec file."""

    bfabric: BfabricAppSpec
    versions: list[AppVersion]

    @field_validator("versions", mode="after")
    def no_duplicate_versions(cls, values: list[AppVersion]) -> list[AppVersion]:
        """Validates that there are no duplicate versions in the app spec."""
        counts = Counter([v.version for v in values])
        duplicates = [version for version, count in counts.items() if count > 1]
        if duplicates:
            raise ValueError(f"Duplicate versions found: {duplicates}")
        return values

    @classmethod
    def load_yaml(cls, app_yaml: Path, app_id: int | str, app_name: str) -> AppSpec:
        """Loads the app versions from the provided YAML file and evaluates the templates."""
        app_spec_file = AppSpecTemplate.model_validate(yaml.safe_load(app_yaml.read_text()))
        return app_spec_file.evaluate(app_id=int(app_id), app_name=str(app_name))

    @property
    def available_versions(self) -> set[str]:
        """The available versions of the app."""
        return {version.version for version in self.versions}

    def __contains__(self, version: str) -> bool:
        return version in self.available_versions

    def __getitem__(self, version: str) -> AppVersion | None:
        """Returns the app version with the provided version number or None if it does not exist."""
        for app_version in self.versions:
            if app_version.version == version:
                return app_version
        return None



from __future__ import annotations

from typing import TYPE_CHECKING

import yaml

if TYPE_CHECKING:
    from pathlib import Path


def write_chunks_file(out_dir: Path, chunks: list[Path]) -> None:
    """Writes the list of chunk paths to the output directory's 'chunks.yml'."""
    out_dir.mkdir(exist_ok=True, parents=True)
    with (out_dir / "chunks.yml").open("w") as f:
        data = {"chunks": [str(chunk) for chunk in chunks]}
        yaml.safe_dump(data, f)



from __future__ import annotations

from typing import TYPE_CHECKING

from bfabric.experimental.workunit_definition import WorkunitDefinition
from bfabric_app_runner.specs.app.app_spec import AppSpec, BfabricAppSpec

if TYPE_CHECKING:
    from pathlib import Path
    from bfabric import Bfabric
    from bfabric_app_runner.specs.app.app_version import AppVersion


def _get_application_version_str(workunit_definition: WorkunitDefinition) -> str:
    """Returns the application version string from the workunit definition."""
    # TODO this should be more generic in the future about the key for the app version (should be handled in AppSpec)
    # TODO logic to define "latest" version (should also be handled in AppSpec)
    if "application_version" not in workunit_definition.execution.raw_parameters:
        msg = "The workunit definition does not specify an application version."
        raise ValueError(msg)
    return workunit_definition.execution.raw_parameters["application_version"]


def _resolve_app(versions: AppSpec, workunit_definition: WorkunitDefinition) -> AppVersion:
    """Resolves the app version to use for the provided workunit definition."""
    version_str = _get_application_version_str(workunit_definition=workunit_definition)
    if version_str not in versions or versions[version_str] is None:
        msg = (
            f"application_version '{version_str}' is not defined in the app spec,\n"
            f" available versions: {sorted(versions.available_versions)}"
        )
        raise ValueError(msg)
    return versions[version_str]


def load_workunit_information(
    app_spec: Path, client: Bfabric, work_dir: Path, workunit_ref: int | Path
) -> tuple[AppVersion, BfabricAppSpec, Path]:
    """Loads the app version and workunit definition from the provided app spec and workunit reference.

    :param app_spec: Path to the app spec file.
    :param client: The B-Fabric client to use for resolving the workunit.
    :param work_dir: Path to the work directory.
    :param workunit_ref: Reference to the workunit (ID or YAML file path).
    :return app_version: The app version to use.
    :return bfabric_app_spec: The B-Fabric app spec, which specifies the app runner version to use.
    :return workunit_ref: The path to the workunit definition file. If the workunit_ref was already a path, it will be
        returned as is, otherwise the file will be created in the work directory.
    """
    workunit_definition, workunit_ref = _resolve_workunit_definition(client, work_dir, workunit_ref)
    app_spec = AppSpec.load_yaml(
        app_spec,
        app_id=workunit_definition.registration.application_id,
        app_name=workunit_definition.registration.application_name,
    )
    app_version = _resolve_app(versions=app_spec, workunit_definition=workunit_definition)
    bfabric_app_spec = app_spec.bfabric
    return app_version, bfabric_app_spec, workunit_ref


def _resolve_workunit_definition(
    client: Bfabric, work_dir: Path, workunit_ref: int | Path
) -> tuple[WorkunitDefinition, Path]:
    """Load the specified workunit definition.

    If the workunit_definition.yml file exists in the work_dir, it will be used, otherwise, it will be created.
    :param client: The B-Fabric client to use for resolving the workunit.
    :param work_dir: The directory where the workunit definition file will be created or found.
    :param workunit_ref: Reference to the workunit (ID or YAML file path).
    :return: A tuple containing the WorkunitDefinition and the path to the workunit definition file.
    """
    workunit_definition_file = work_dir / "workunit_definition.yml"
    workunit_definition = WorkunitDefinition.from_ref(workunit_ref, client, cache_file=workunit_definition_file)
    return workunit_definition, workunit_definition_file



from __future__ import annotations

from enum import Enum
from typing import Annotated

from pydantic import BaseModel, field_validator, Field, AliasChoices


class SlurmWorkunitSpecialStrings(Enum):
    default = "[default]"
    auto = "[auto]"

    @classmethod
    def parse(cls, string: str) -> SlurmWorkunitSpecialStrings | None:
        if string == cls.default.value:
            return cls.default
        elif string == cls.auto.value:
            return cls.auto
        else:
            return None


class SlurmWorkunitParams(BaseModel):
    partition: Annotated[
        SlurmWorkunitSpecialStrings | str, Field(validation_alias=AliasChoices("partition", "--partition"))
    ] = SlurmWorkunitSpecialStrings.default
    nodelist: Annotated[
        SlurmWorkunitSpecialStrings | str, Field(validation_alias=AliasChoices("nodelist", "--nodelist"))
    ] = SlurmWorkunitSpecialStrings.default
    mem: Annotated[SlurmWorkunitSpecialStrings | str, Field(validation_alias=AliasChoices("mem", "--mem"))] = (
        SlurmWorkunitSpecialStrings.default
    )
    time: Annotated[SlurmWorkunitSpecialStrings | str, Field(validation_alias=AliasChoices("time", "--time"))] = (
        SlurmWorkunitSpecialStrings.default
    )

    @classmethod
    def _parse_string(cls, value: str | SlurmWorkunitSpecialStrings) -> SlurmWorkunitSpecialStrings | str:
        if isinstance(value, SlurmWorkunitSpecialStrings):
            return value
        parsed = SlurmWorkunitSpecialStrings.parse(value)
        if parsed is not None:
            return parsed
        return value

    @field_validator("partition", "nodelist", "mem", "time", mode="before")
    def validate_special_string(cls, value: SlurmWorkunitSpecialStrings | str) -> SlurmWorkunitSpecialStrings | str:
        return cls._parse_string(value)

    def _get_field(self, field_name: str) -> str | None:
        """Returns the value to be used for the field, or None if the default should be used (when merging configs)."""
        value = getattr(self, field_name)
        if isinstance(value, str):
            return value
        if value == SlurmWorkunitSpecialStrings.default:
            return None
        else:
            raise NotImplementedError(f"Currently unsupported value for {field_name}: {value}")

    def as_dict(self) -> dict[str, str | None]:
        """Returns a dictionary with the values that should be passed to sbatch."""
        field_mapping = {
            "partition": "--partition",
            "nodelist": "--nodelist",
            "mem": "--mem",
            "time": "--time",
        }

        values = {}
        for field_name, flag in field_mapping.items():
            value = self._get_field(field_name)
            if value is not None:
                values[flag] = value

        return values



from __future__ import annotations
from typing import Literal, TYPE_CHECKING

from pydantic import BaseModel, ConfigDict

from bfabric_app_runner.specs.common_types import RelativeFilePath  # noqa: TC001

if TYPE_CHECKING:
    from bfabric import Bfabric


class BfabricOrderFastaSpec(BaseModel):
    model_config = ConfigDict(extra="forbid")
    type: Literal["bfabric_order_fasta"] = "bfabric_order_fasta"

    id: int
    entity: Literal["workunit", "order"]
    filename: RelativeFilePath
    required: bool = False

    def resolve_filename(self, client: Bfabric) -> str:
        return self.filename



from __future__ import annotations

from typing import TYPE_CHECKING

from bfabric_app_runner.inputs.resolve.resolved_inputs import ResolvedDirectory
from bfabric_app_runner.specs.inputs.file_spec import FileSourceSsh, FileSourceSshValue

from bfabric.entities import Resource, Storage

if TYPE_CHECKING:
    from bfabric import Bfabric
    from bfabric_app_runner.specs.inputs.bfabric_resource_archive_spec import BfabricResourceArchiveSpec


class ResolveBfabricResourceArchiveSpecs:
    def __init__(self, client: Bfabric) -> None:
        self._client = client

    def __call__(self, specs: list[BfabricResourceArchiveSpec]) -> list[ResolvedDirectory]:
        """Convert resource archive specifications to directory specifications."""
        if not specs:
            return []

        # Fetch all resources and their storage information in bulk
        resource_ids = [spec.id for spec in specs]
        resources = Resource.find_all(ids=resource_ids, client=self._client)
        storage_ids = sorted({resource["storage"]["id"] for resource in resources.values()})
        storages = Storage.find_all(ids=storage_ids, client=self._client)

        # Create the directory specs
        result = []
        for resource_id, spec in zip(resource_ids, specs):
            resource = resources[resource_id]
            storage = storages[resource["storage"]["id"]]
            result.append(self._get_directory_spec(spec=spec, resource=resource, storage=storage))

        return result

    def _get_directory_spec(
        self, spec: BfabricResourceArchiveSpec, resource: Resource, storage: Storage
    ) -> ResolvedDirectory:
        return ResolvedDirectory(
            source=FileSourceSsh(
                ssh=FileSourceSshValue(host=storage["host"], path=f"{storage['basepath']}{resource['relativepath']}")
            ),
            filename=spec.filename,
            extract=spec.extract,
            include_patterns=spec.include_patterns,
            exclude_patterns=spec.exclude_patterns,
            strip_root=spec.strip_root,
            checksum=resource["filechecksum"] if spec.check_checksum else None,
        )



from pathlib import Path

import yaml
from rich.pretty import pprint

from bfabric_app_runner.specs.app.app_spec import AppSpecTemplate, AppSpec
from bfabric_app_runner.specs.inputs_spec import InputsSpec
from bfabric_app_runner.specs.outputs_spec import OutputsSpec


def cmd_validate_app_spec_template(yaml_file: Path) -> None:
    """Validate an app spec file."""
    app_spec_file = AppSpecTemplate.model_validate(yaml.safe_load(yaml_file.read_text()))
    pprint(app_spec_file)


def cmd_validate_app_spec(app_yaml: Path, app_id: int = 0, app_name: str = "x") -> None:
    """Validates the app versions by expanding the relevant config info."""
    versions = AppSpec.load_yaml(app_yaml, app_id=app_id, app_name=app_name)
    pprint(versions)


def cmd_validate_inputs_spec(yaml_file: Path) -> None:
    """Validate an inputs spec file."""
    inputs_spec = InputsSpec.model_validate(yaml.safe_load(yaml_file.read_text()))
    pprint(inputs_spec)


def cmd_validate_outputs_spec(yaml_file: Path) -> None:
    """Validate an outputs spec file."""
    outputs_spec = OutputsSpec.model_validate(yaml.safe_load(yaml_file.read_text()))
    pprint(outputs_spec)



"""
B-Fabric App Runner - Execute bioinformatics applications on your B-Fabric workunits.

**For automated systems:**
Use ``bfabric-app-runner run workunit`` for end-to-end workunit execution.

**For development and testing:**

1. Create a workunit folder: ``bfabric-app-runner prepare workunit``
2. Navigate to the folder and run ``make help`` for available commands

**Options:**

- Add ``--read-only`` to prevent writes to your B-Fabric instance during testing
- Add ``--use-external-runner`` if ``bfabric-app-runner`` is available externally (otherwise requires ``uv``)

The generated Makefile uses ``bfabric-app-runner action`` commands internally.
"""

from __future__ import annotations

import importlib.metadata

import cyclopts

from bfabric_app_runner.cli.cmd_action import (
    cmd_action_inputs,
    cmd_action_outputs,
    cmd_action_process,
    cmd_action_run_all,
    cmd_action_dispatch,
)
from bfabric_app_runner.cli.cmd_prepare import cmd_prepare_workunit
from bfabric_app_runner.cli.cmd_run import cmd_run_workunit
from bfabric_app_runner.cli.inputs import cmd_inputs_prepare, cmd_inputs_clean, cmd_inputs_list, cmd_inputs_check
from bfabric_app_runner.cli.outputs import cmd_outputs_register, cmd_outputs_register_single_file
from bfabric_app_runner.cli.validate import (
    cmd_validate_inputs_spec,
    cmd_validate_outputs_spec,
    cmd_validate_app_spec,
    cmd_validate_app_spec_template,
)


package_version = importlib.metadata.version("bfabric_app_runner")
app = cyclopts.App(
    help=__doc__,
    version=package_version,
)

groups = {
    "Running Apps": cyclopts.Group("Running Apps", sort_key=1),
    "Developer Tools": cyclopts.Group("Developer Tools", sort_key=2),
}


cmd_inputs = cyclopts.App("inputs", help="Prepare input files for an app.")
cmd_inputs.command(cmd_inputs_check, name="check")
cmd_inputs.command(cmd_inputs_clean, name="clean")
cmd_inputs.command(cmd_inputs_list, name="list")
cmd_inputs.command(cmd_inputs_prepare, name="prepare")
cmd_inputs.group = groups["Developer Tools"]
app.command(cmd_inputs)

cmd_outputs = cyclopts.App("outputs", help="Register output files of an app.")
cmd_outputs.command(cmd_outputs_register, name="register")
cmd_outputs.command(cmd_outputs_register_single_file, name="register-single-file")
cmd_outputs.group = groups["Developer Tools"]
app.command(cmd_outputs)

cmd_validate = cyclopts.App("validate", help="Validate yaml files.")
cmd_validate.command(cmd_validate_app_spec, name="app-spec")
cmd_validate.command(cmd_validate_app_spec_template, name="app-spec-template")
cmd_validate.command(cmd_validate_inputs_spec, name="inputs-spec")
cmd_validate.command(cmd_validate_outputs_spec, name="outputs-spec")
cmd_validate.group = groups["Developer Tools"]
app.command(cmd_validate)

cmd_action = cyclopts.App("action", help="Executes an action of a prepared workunit")
cmd_action.command(cmd_action_run_all, name="run-all")
cmd_action.command(cmd_action_dispatch, name="dispatch")
cmd_action.command(cmd_action_inputs, name="inputs")
cmd_action.command(cmd_action_process, name="process")
cmd_action.command(cmd_action_outputs, name="outputs")
cmd_action.group = groups["Developer Tools"]
app.command(cmd_action)

cmd_prepare = cyclopts.App(name="prepare", help="Prepare a workunit for execution.")
cmd_prepare.command(cmd_prepare_workunit, name="workunit")
cmd_prepare.group = groups["Running Apps"]
app.command(cmd_prepare)

cmd_run = cyclopts.App(name="run", help="Run an app end-to-end.")
cmd_run.command(cmd_run_workunit, name="workunit")
cmd_run.group = groups["Running Apps"]
app.command(cmd_run)

if __name__ == "__main__":
    app()



from __future__ import annotations

import re
import secrets
from typing import Any

from loguru import logger
from mako.template import Template
from pydantic import BaseModel, field_validator, Field


class VariablesApp(BaseModel):
    """App specific variables that can be used in our config templates.

    They will be available as `${app.id}` etc. in the config files.
    """

    id: int
    name: str
    version: str

    @field_validator("name", mode="before")
    def validate_safe_name(cls, value: str) -> str:
        characters = re.compile(r"[^a-zA-Z0-9_-]+")
        return characters.sub("_", value)


class VariablesWorkunit(BaseModel):
    id: int
    file_token: str = Field(default_factory=lambda: secrets.token_hex(16))


class Variables(BaseModel):
    """Variables that can be used in our config templates."""

    app: VariablesApp
    workunit: VariablesWorkunit | None

    def as_dict(self) -> dict[str, VariablesApp | VariablesWorkunit | None]:
        return {"app": self.app, "workunit": self.workunit}


def interpolate_config_strings(data: Any, variables: Variables | dict[str, Any]) -> Any:
    """Recursively evaluates all strings in a data structure with Mako templates.

    This will not evaluate Mako templates in the YAML file itself, only in the individual strings.
    Since the current behavior is a subset of evaluating all strings in the YAML file, we could extend this later
    if it becomes necessary. However, it has the risk of making the config files more complex and should be avoided
    if possible.

    :param data: Any Python data structure (dict, list, str, etc.)
    :param variables: Template variables and their values
    :return: The data structure with all strings evaluated
    """
    variables = Variables.model_validate(variables) if isinstance(variables, dict) else variables
    logger.debug(f"Interpolating config strings with variables: {variables}")

    if isinstance(data, dict):
        return {key: interpolate_config_strings(value, variables) for key, value in data.items()}
    elif isinstance(data, list):
        return [interpolate_config_strings(item, variables) for item in data]
    elif isinstance(data, str):
        return str(Template(data).render(**variables.as_dict()))
    else:
        return data



from pathlib import Path

from loguru import logger

from bfabric import Bfabric
from bfabric.entities import Dataset
from bfabric.experimental.workunit_definition import WorkunitDefinition
from bfabric_app_runner.dispatch.generic import write_chunks_file


class DispatchSingleDatasetFlow:
    def __init__(self, client: Bfabric, out_dir: Path) -> None:
        self._client = client
        self._out_dir = out_dir

    def dispatch_job(self, dataset: Dataset, workunit: WorkunitDefinition) -> Path:
        raise NotImplementedError

    def dispatch_workunit(self, workunit: WorkunitDefinition) -> None:
        if not workunit.execution.dataset:
            logger.error("No dataset found in workunit.")
            return

        dataset = Dataset.find(id=workunit.execution.dataset, client=self._client)
        path = self.dispatch_job(dataset=dataset, workunit=workunit)
        write_chunks_file(out_dir=self._out_dir, chunks=[path])



from __future__ import annotations

import enum
from pathlib import Path  # noqa: TCH003
from typing import Literal, Annotated

import yaml
from pydantic import BaseModel, ConfigDict, Field, model_validator


class UpdateExisting(enum.Enum):
    NO = "no"
    IF_EXISTS = "if_exists"
    REQUIRED = "required"


class CopyResourceSpec(BaseModel):
    model_config = ConfigDict(extra="forbid")
    type: Literal["bfabric_copy_resource"] = "bfabric_copy_resource"

    local_path: Path
    """The local path to the file to be copied."""

    store_entry_path: Path
    """The path to the storage entry in the storage folder."""

    store_folder_path: Path | None = None
    """The storage folder will be determined by the default rule, but can be specified if needed."""

    # TODO these need to be implemented properly (e.g. do not scp too early), and tested in integration tests
    update_existing: UpdateExisting = UpdateExisting.IF_EXISTS

    protocol: Literal["scp"] = "scp"


class SaveDatasetSpec(BaseModel):
    model_config = ConfigDict(extra="forbid")
    type: Literal["bfabric_dataset"] = "bfabric_dataset"

    # TODO this will currently fail if the workunit already has an output dataset -> needs to be handled as well
    local_path: Path
    separator: str
    name: str | None = None
    has_header: bool = True
    invalid_characters: str = ""


class SaveLinkSpec(BaseModel):
    """Saves a link to the workunit, or, if desired to an arbitrary entity of type entity_type with id entity_id."""

    model_config = ConfigDict(extra="forbid")
    type: Literal["bfabric_link"] = "bfabric_link"
    name: str
    """The name of the link."""
    url: str
    """The URL of the link."""
    entity_type: str = "Workunit"
    """The type of the entity that will be linked."""
    entity_id: int | None = None
    """The ID of the entity that will be linked."""
    update_existing: UpdateExisting = UpdateExisting.IF_EXISTS
    """Behavior, if a link with the same name already exists."""

    @model_validator(mode="after")
    def require_entity_id_if_not_workunit(self) -> SaveLinkSpec:
        if self.entity_type != "Workunit" and self.entity_id is None:
            raise ValueError("entity_id must be provided if entity_type is not 'Workunit'")
        return self


SpecType = Annotated[CopyResourceSpec | SaveDatasetSpec | SaveLinkSpec, Field(discriminator="type")]


class OutputsSpec(BaseModel):
    model_config = ConfigDict(extra="forbid")
    outputs: list[SpecType]

    @classmethod
    def read_yaml(cls, path: Path) -> list[SpecType]:
        model = cls.model_validate(yaml.safe_load(path.read_text()))
        return model.outputs

    @classmethod
    def write_yaml(cls, specs: list[SpecType], path: Path) -> None:
        model = cls.model_validate(dict(outputs=specs))
        path.write_text(yaml.dump(model.model_dump(mode="json")))



from pathlib import Path
from typing import Protocol, Any

import cyclopts
import pandera.polars as pa
import polars as pl
import yaml
from pandera import Field
from pandera.typing.polars import DataFrame

from bfabric import Bfabric
from bfabric.entities import Resource
from bfabric.experimental.workunit_definition import WorkunitDefinition
from bfabric.utils.cli_integration import use_client


class InputTable(pa.DataFrameModel):
    resource_id: int
    filename: str
    relativepath: str


class OutputTable(pa.DataFrameModel):
    class Config:
        coerce = True
        add_missing_columns = True

    resource_id: int
    filename: str = Field(nullable=True)
    task: str = Field(default=pl.lit("work"), str_matches=r"^\w+$")


class DispatchResource(Protocol):
    """A callable that dispatches resources for processing.

    It receives a table of all available resources and returns a table of
    resources to be processed by the workunit.

    If you check the difference between `InputTable` and `OutputTable`, you can see that `OutputTable` has an additional
    field `task` which is set to the string "work". If your workunit entails separate tasks to be computed, you can
    assign different values to the `task` field to differentiate between them.
    """

    def __call__(
        self, resources_df: DataFrame[InputTable], workunit_definition: WorkunitDefinition
    ) -> DataFrame[OutputTable] | tuple[DataFrame[OutputTable], list[dict[str, Any]]]:
        """Return a table of resources to be processed by the workunit.

        :param resources_df: Input DataFrame to be processed, of InputTable schema
        :param workunit_definition: The workunit definition, if you need additional information
        :return:
          - Output DataFrame to be processed, of OutputTable schema
          - Optional: a list of additional input specs to add to the inputs.yml file (if multiple tasks are generated,
            this will be added to each task's inputs.yml file)
        """
        ...


class ResourceDispatcher:
    def dispatch(
        self,
        workunit_definition: WorkunitDefinition,
        work_dir: Path,
        resource_strategy: DispatchResource,
        *,
        client: Bfabric,
    ) -> None:
        """Dispatches a resource flow workunit with the specified resource strategy.

        :param workunit_definition: The workunit definition to dispatch.
        :param work_dir: The workunit working directory.
        :param resource_strategy: The resource dispatch strategy.
        :param client: The B-Fabric client to use.
        """
        extra_inputs, requested_table, tasks = self._determine_requested_inputs(
            client=client, resource_strategy=resource_strategy, workunit_definition=workunit_definition
        )
        for task in tasks:
            self._create_task_dir(
                extra_inputs=extra_inputs, requested_table=requested_table, task=task, work_dir=work_dir
            )

        # Generate the `chunks.yml` file
        with (work_dir / "chunks.yml").open("w") as f:
            yaml.safe_dump({"chunks": tasks}, f)

    @staticmethod
    def _build_input_resources_df(resource_ids: list[int], client: Bfabric) -> DataFrame[InputTable]:
        """Creates the InputTable DataFrame from the list of resource IDs."""
        resources = Resource.find_all(resource_ids, client=client).values()
        if not resources:
            raise ValueError("No resources to dispatch")
        attributes = [
            {
                "resource_id": resource["id"],
                "filename": Path(resource["relativepath"]).name,
                "relativepath": resource["relativepath"],
            }
            for resource in resources
        ]
        df = pl.DataFrame(attributes)
        return InputTable.validate(df)

    @staticmethod
    def _build_inputs_spec(inputs: DataFrame[OutputTable]) -> dict[str, list[dict]]:
        """Builds input specs from the OutputTable DataFrame."""
        specs = []
        for row in inputs.iter_rows(named=True):
            specs.append({"type": "bfabric_resource", "id": row["resource_id"], "filename": row["filename"]})
        return {"inputs": specs}

    @staticmethod
    def _evaluate_strategy(
        input_resources_df: DataFrame[InputTable],
        workunit_definition: WorkunitDefinition,
        resource_strategy: DispatchResource,
    ) -> tuple[DataFrame[OutputTable], list[dict[str, Any]]]:
        result = resource_strategy(input_resources_df, workunit_definition)
        if isinstance(result, tuple):
            return OutputTable.validate(result[0]), result[1]
        return OutputTable.validate(result), []

    def _determine_requested_inputs(
        self, client: Bfabric, resource_strategy: DispatchResource, workunit_definition: WorkunitDefinition
    ) -> tuple[list[dict[str, Any]], DataFrame[OutputTable], list[str]]:
        input_resources_df = self._build_input_resources_df(
            resource_ids=workunit_definition.execution.resources, client=client
        )
        requested_table, extra_inputs = self._evaluate_strategy(
            input_resources_df=input_resources_df,
            workunit_definition=workunit_definition,
            resource_strategy=resource_strategy,
        )
        tasks = sorted(requested_table["task"].unique())
        return extra_inputs, requested_table, tasks

    def _create_task_dir(
        self, extra_inputs: list[dict[str, Any]], requested_table: DataFrame[OutputTable], task: str, work_dir: Path
    ) -> None:
        """Creates a task directory with the specified inputs."""
        task_dir = work_dir / task
        task_dir.mkdir(exist_ok=True, parents=True)

        # Generate the task `inputs.yml` file
        task_inputs = requested_table.filter(pl.col("task") == pl.lit(task))
        inputs_spec = self._build_inputs_spec(task_inputs)
        inputs_spec["inputs"].extend(extra_inputs)
        with (task_dir / "inputs.yml").open("w") as f:
            yaml.safe_dump(inputs_spec, f)


class ResourceDispatcherCLI:
    """A CLI interface for dispatching resources to workunits."""

    def __init__(self, resource_strategy: DispatchResource) -> None:
        self._resource_strategy = resource_strategy
        self._app = cyclopts.App()
        self._app.default(self.main)

    def run(self) -> None:
        """Runs the CLI application."""
        self._app()

    @use_client
    def main(self, workunit_ref: Path | int, out_dir: Path, *, client: Bfabric) -> None:
        """Dispatches resources to a workunit."""
        workunit_definition = WorkunitDefinition.from_ref(workunit_ref, client=client)
        dispatcher = ResourceDispatcher()
        dispatcher.dispatch(
            workunit_definition=workunit_definition,
            work_dir=out_dir,
            resource_strategy=self._resource_strategy,
            client=client,
        )



from __future__ import annotations

from typing import Any, TYPE_CHECKING

from pydantic import BaseModel, ConfigDict, model_validator

from bfabric.entities import Resource, Dataset
from bfabric_app_runner.dispatch.generic import write_chunks_file
from bfabric_app_runner.dispatch.resource_flow import get_resource_flow_input_resources

if TYPE_CHECKING:
    from pathlib import Path
    from bfabric import Bfabric
    from bfabric.experimental.workunit_definition import WorkunitDefinition


class ConfigResourceFlow(BaseModel):
    model_config = ConfigDict(extra="forbid")
    filter_suffix: str | None = None


class ConfigDatasetFlow(BaseModel):
    model_config = ConfigDict(extra="forbid")
    resource_column: str = "Imzml"
    param_columns: list[tuple[str, str]] = [("PanelDataset", "mass_list_id")]


class ConfigDispatchIndividualResources(BaseModel):
    resource_flow: ConfigResourceFlow | None
    dataset_flow: ConfigDatasetFlow | None

    @model_validator(mode="after")
    def check_at_least_one_flow(self) -> ConfigDispatchIndividualResources:
        if self.resource_flow is None and self.dataset_flow is None:
            raise ValueError("either resource_flow or dataset_flow must be provided")
        return self


def config_msi_imzml() -> ConfigDispatchIndividualResources:
    """Returns the configuration for dispatching MSI imzML datasets to chunks.

    These apps allow both being run with a list of input `.imzML` resource files, or a dataset which contains a column
    `Imzml` with the resource IDs and a column `PanelDataset` with the dataset IDs.

    Note: In the future the specifics of this might be adapted to allow e.g. `.imzML.7z` files or similar.
    """
    return ConfigDispatchIndividualResources(
        resource_flow=ConfigResourceFlow(filter_suffix=".imzML"),
        dataset_flow=ConfigDatasetFlow(resource_column="Imzml", param_columns=[("PanelDataset", "mass_list_id")]),
    )


class DispatchIndividualResources:
    """Dispatches jobs on individual resources specified in the workunit."""

    def __init__(self, client: Bfabric, config: ConfigDispatchIndividualResources, out_dir: Path) -> None:
        self._client = client
        self._config = config
        self._out_dir = out_dir

    def dispatch_job(self, resource: Resource, params: dict[str, Any]) -> Path:
        # TODO it is not clear yet if this is the best approach to handle this by inheritance here,
        #      I would sort of prefer to decouple this but it's not fully clear how to do it and it would not always
        #      be possible to do this as generically
        # -> maybe in an initial version we can handle it with python code, but in the future revisit if it might make
        #    more sense to be even more generic there
        raise NotImplementedError

    def dispatch_workunit(self, definition: WorkunitDefinition) -> None:
        params = definition.execution.raw_parameters
        if definition.execution.resources:
            paths = self._dispatch_jobs_resource_flow(definition, params)
        elif definition.execution.dataset:
            paths = self._dispatch_jobs_dataset_flow(definition, params)
        else:
            raise ValueError("either dataset or resources must be provided")
        write_chunks_file(out_dir=self._out_dir, chunks=paths)

    def _dispatch_jobs_resource_flow(self, definition: WorkunitDefinition, params: dict[str, Any]) -> list[Path]:
        """Returns the individual jobs for a resource flow workunit and returns the paths of the task folders."""
        config = self._config.resource_flow
        if config is None:
            raise ValueError("resource_flow is not configured")
        resources = get_resource_flow_input_resources(
            client=self._client, definition=definition, filter_suffix=config.filter_suffix
        )
        return [self.dispatch_job(resource=resource, params=params) for resource in resources]

    def _dispatch_jobs_dataset_flow(self, definition: WorkunitDefinition, params: dict[str, Any]) -> list[Path]:
        config = self._config.dataset_flow
        if config is None:
            raise ValueError("dataset_flow is not configured")
        dataset = Dataset.find(id=definition.execution.dataset, client=self._client)
        if dataset is None:
            msg = f"Dataset with id {definition.execution.dataset} not found"
            raise ValueError(msg)
        dataset_df = dataset.to_polars()
        resources = Resource.find_all(ids=dataset_df[config.resource_column].unique().to_list(), client=self._client)
        paths = []
        for row in dataset_df.iter_rows(named=True):
            resource_id = int(row[config.resource_column])
            row_params = {name: row[dataset_name] for dataset_name, name in config.param_columns}
            paths.append(self.dispatch_job(resource=resources[resource_id], params=params | row_params))
        return paths



from __future__ import annotations

from pathlib import Path
from collections import defaultdict
from typing import Annotated, Literal, Self

from bfabric_app_runner.specs.common_types import RelativeFilePath  # noqa: TC001
from bfabric_app_runner.specs.inputs.file_spec import FileSourceSsh, FileSourceLocal  # noqa: TC001
from pydantic import BaseModel, Field, model_validator


class ResolvedFile(BaseModel):
    type: Literal["resolved_file"] = "resolved_file"
    filename: RelativeFilePath
    source: FileSourceSsh | FileSourceLocal
    # TODO later, we should consider if it would make sense to split linking into a separate class
    link: bool
    checksum: str | None


class ResolvedStaticFile(BaseModel):
    type: Literal["resolved_static_file"] = "resolved_static_file"
    filename: RelativeFilePath
    content: str | bytes


class ResolvedDirectory(BaseModel):
    type: Literal["resolved_directory"] = "resolved_directory"
    filename: RelativeFilePath
    source: FileSourceSsh | FileSourceLocal
    extract: None | Literal["zip"] = None
    include_patterns: list[str] = []
    exclude_patterns: list[str] = []
    strip_root: bool = False
    checksum: str | None = None


ResolvedInput = ResolvedFile | ResolvedStaticFile | ResolvedDirectory


class ResolvedInputs(BaseModel):
    files: list[Annotated[ResolvedInput, Field(discriminator="type")]]

    @model_validator(mode="after")
    def no_duplicates(self) -> Self:
        # Build efficient data structures in a single pass
        filename_counts = defaultdict(int)
        directories = set()
        all_paths = set()

        for file in self.files:
            filename = file.filename
            filename_counts[filename] += 1
            all_paths.add(filename)

            if isinstance(file, ResolvedDirectory):
                directories.add(filename)

        # Check for exact duplicates - O(1) lookup
        duplicates = [name for name, count in filename_counts.items() if count > 1]
        if duplicates:
            msg = f"Duplicate filenames in resolved inputs: {', '.join(sorted(duplicates))}"
            raise ValueError(msg)

        # Check for "." conflicts - O(1) lookup
        if "." in all_paths and len(all_paths) > 1:
            msg = "Current directory '.' cannot coexist with other files"
            raise ValueError(msg)

        # Check for directory/file conflicts - O(d * n), where d = number of directories, n = total files
        for directory in directories:
            dir_path = Path(directory)
            for path in all_paths:
                file_path = Path(path)
                if path != directory and file_path.is_relative_to(dir_path):
                    msg = (
                        f"Path '{path}' conflicts with directory '{directory}'"
                        f" (would be inside extracted directory)"
                    )
                    raise ValueError(msg)

        return self

    def apply_filter(self, filter_files: list[str]) -> Self:
        """Returns a new instance with only the files that are in the filter_files list."""
        return type(self)(files=[file for file in self.files if file.filename in filter_files])



from typing import Literal, Annotated

import cyclopts
import shlex
import subprocess
from loguru import logger
from pathlib import Path
from bfabric import Bfabric
from bfabric.entities import ExternalJob, Workunit
from bfabric.utils.cli_integration import use_client
from bfabric_app_runner.bfabric_integration.wrapper.wrap_app_yaml_template import WrapAppYamlTemplate
from bfabric_app_runner.bfabric_integration.submitter.config.slurm_params import evaluate_slurm_parameters
from bfabric_app_runner.bfabric_integration.submitter.slurm_job_template import SlurmJobTemplate

app = cyclopts.App()


JobIdParam = Annotated[int, cyclopts.Parameter(name="-j", help="The externaljob ID to process.")]


@app.command()
def wrapper_creator(j: JobIdParam) -> None:
    """Wrapper creator for simple_submitter."""
    logger.info("Wrapper creator called with job ID:", j)


@app.command()
@use_client
def submitter(
    config_path: Path,
    j: JobIdParam,
    *,
    entity_type: Literal["externaljob", "workunit"] = "externaljob",
    client: Bfabric,
) -> None:
    """Submitter implementation for simple_submitter."""
    if entity_type == "externaljob":
        # Find the workunit to process
        external_job = ExternalJob.find(id=j, client=client)
        workunit = external_job.workunit
        if workunit is None:
            raise RuntimeError(f"External job {j} does not belong to a workunit (or it was deleted).")
    else:
        workunit = Workunit.find(id=j, client=client)

    _submit_workunit(workunit=workunit, config_path=config_path)


def _submit_workunit(workunit: Workunit, config_path: Path) -> None:
    """Submit a workunit to the bfabric app runner."""
    logger.info(f"Submitting workunit {workunit.id}.")
    slurm_params = evaluate_slurm_parameters(config_yaml_path=config_path, workunit=workunit)

    # Create the wrapped script
    wrap_app_yaml_template_params = WrapAppYamlTemplate.Params.extract_workunit(
        workunit, scratch_root=slurm_params.scratch_root
    )
    wrap_app_yaml_template = WrapAppYamlTemplate(
        params=wrap_app_yaml_template_params, path=WrapAppYamlTemplate.default_path()
    )
    wrapped_script = wrap_app_yaml_template.render_string()

    # Create the slurm executable
    slurm_job_template = SlurmJobTemplate(
        params=slurm_params,
        workunit_id=workunit.id,
        wrapped_script=wrapped_script,
        path=SlurmJobTemplate.default_path(),
    )
    slurm_script = slurm_job_template.render_string()

    # Ensure log dir exists
    if "--output" in slurm_params.sbatch_params:
        log_dir = Path(slurm_params.sbatch_params["--output"]).parent
        log_dir.mkdir(parents=True, exist_ok=True)

    # Write this slurm script to the appropriate location
    slurm_script_path = slurm_params.job_script
    slurm_script_path.parent.mkdir(exist_ok=True)
    slurm_script_path.write_text(slurm_script)

    # Submit the slurm script and be done.
    cmd = ["sbatch", str(slurm_script_path)]
    logger.info(f"Running slurm submit: {shlex.join(cmd)}")
    subprocess.run(cmd, check=True)


if __name__ == "__main__":
    app()



import warnings
from pathlib import Path
from typing import Any

from loguru import logger

from bfabric import Bfabric
from bfabric.entities import WorkflowTemplateStep, WorkflowTemplate, WorkflowStep
from bfabric.experimental.entity_lookup_cache import EntityLookupCache
from bfabric.experimental.workunit_definition import WorkunitDefinition
from bfabric_app_runner.actions.types import (
    ActionDispatch,
    ActionInputs,
    ActionProcess,
    ActionOutputs,
    ActionRun,
)
from bfabric_app_runner.app_runner.resolve_app import load_workunit_information
from bfabric_app_runner.app_runner.runner import ChunksFile
from bfabric_app_runner.app_runner.runner import Runner
from bfabric_app_runner.inputs.prepare.prepare_folder import prepare_folder
from bfabric_app_runner.output_registration import register_outputs


def execute_dispatch(action: ActionDispatch, client: Bfabric) -> None:
    """Executes a dispatch action."""
    app_version, bfabric_app_spec, workunit_definition_path = load_workunit_information(
        app_spec=action.app_ref, client=client, work_dir=action.work_dir, workunit_ref=action.workunit_ref
    )
    workunit_definition_mtime = workunit_definition_path.stat().st_mtime
    workunit_definition = WorkunitDefinition.from_yaml(workunit_definition_path)

    with EntityLookupCache.enable():
        runner = Runner(spec=app_version, client=client, ssh_user=None)
        runner.run_dispatch(workunit_ref=workunit_definition_path, work_dir=action.work_dir)

    new_mtime = workunit_definition_path.stat().st_mtime
    if workunit_definition_mtime != new_mtime:
        warnings.warn(
            f"The workunit definition file {workunit_definition_path} was modified during dispatch. "
            "This must not happen, and could be promoted to an error in the future. "
            "Please adapt the app, restoring original file now.",
            DeprecationWarning,
        )
        logger.warning("Renaming workunit_definition.yml to workunit_definition.yml.bak, and restoring original.")
        workunit_definition_path.rename(workunit_definition_path.with_suffix(".yml.bak"))
        workunit_definition.to_yaml(workunit_definition_path)

    if not action.read_only:
        # Set the workunit status to processing
        client.save("workunit", {"id": workunit_definition.registration.workunit_id, "status": "processing"})

        # Create a workflowstep template if specified
        if bfabric_app_spec.workflow_template_step_id:
            logger.info(f"Creating workflowstep from template {bfabric_app_spec.workflow_template_step_id}")
            _register_workflow_step(
                workflow_template_step_id=bfabric_app_spec.workflow_template_step_id,
                workunit_definition=workunit_definition,
                client=client,
            )


def execute_run(action: ActionRun, client: Bfabric) -> None:
    """Executes a run action."""
    chunk_dirs = _validate_chunks_list(action.work_dir, action.chunk)
    for chunk_dir_rel in chunk_dirs:
        execute_inputs(action=ActionInputs.from_action_run(action, chunk=str(chunk_dir_rel)), client=client)
        execute_process(action=ActionProcess.from_action_run(action, chunk=str(chunk_dir_rel)), client=client)
        execute_outputs(action=ActionOutputs.from_action_run(action, chunk=str(chunk_dir_rel)), client=client)


def execute_inputs(action: ActionInputs, client: Bfabric) -> None:
    """Executes an inputs action."""
    chunk_dirs = _validate_chunks_list(action.work_dir, action.chunk)
    for chunk_dir_rel in chunk_dirs:
        prepare_folder(
            inputs_yaml=chunk_dir_rel / "inputs.yml",
            target_folder=chunk_dir_rel,
            client=client,
            ssh_user=action.ssh_user,
            filter=action.filter,
            action="prepare",
        )


def execute_process(action: ActionProcess, client: Bfabric) -> None:
    """Executes a process action."""
    chunk_dirs = _validate_chunks_list(action.work_dir, action.chunk)

    for chunk_dir_rel in chunk_dirs:
        app_spec = action.app_ref
        chunk_dir = chunk_dir_rel.resolve()
        app_version, _, workunit_ref = load_workunit_information(
            app_spec=app_spec,
            client=client,
            work_dir=action.work_dir,
            workunit_ref=action.work_dir / "workunit_definition.yml",
        )

        with EntityLookupCache.enable():
            runner = Runner(spec=app_version, client=client, ssh_user=None)
            runner.run_process(chunk_dir=chunk_dir)


def execute_outputs(action: ActionOutputs, client: Bfabric) -> None:
    """Executes an outputs action."""
    chunk_dirs = _validate_chunks_list(action.work_dir, action.chunk)

    for chunk_dir_rel in chunk_dirs:
        # this includes the legacy collect step
        chunk_dir = chunk_dir_rel.resolve()
        app_version, _, workunit_ref = load_workunit_information(
            app_spec=action.app_ref, client=client, work_dir=chunk_dir, workunit_ref=action.workunit_ref
        )
        runner = Runner(spec=app_version, client=client, ssh_user=action.ssh_user)
        runner.run_collect(workunit_ref=action.workunit_ref, chunk_dir=chunk_dir)
        workunit_definition = WorkunitDefinition.from_yaml(path=action.work_dir / "workunit_definition.yml")
        if not action.read_only:
            register_outputs(
                outputs_yaml=chunk_dir / "outputs.yml",
                workunit_definition=workunit_definition,
                client=client,
                ssh_user=action.ssh_user,
                force_storage=action.force_storage,
                reuse_default_resource=True,
            )


def _register_workflow_step(
    workflow_template_step_id: int, workunit_definition: WorkunitDefinition, client: Bfabric
) -> None:
    # Load the workflow template step
    workflow_template_step = WorkflowTemplateStep.find(id=workflow_template_step_id, client=client)
    if not workflow_template_step:
        logger.error(f"Misconfigured {workflow_template_step_id=}, cannot find it in the database.")
        return

    # Find or create the workflow entity
    workflow = _find_or_create_workflow(
        workflow_template=workflow_template_step.workflow_template,
        workunit_definition=workunit_definition,
        client=client,
    )

    # Prepare the data for the workflow step
    workunit_id = workunit_definition.registration.workunit_id
    workflow_step_data = {
        "workflowid": workflow["id"],
        "workflowtemplatestepid": workflow_template_step.id,
        "supervisorid": workunit_definition.registration.user_id,
        "workunitid": workunit_id,
    }
    if workunit_definition.execution.dataset is not None:
        workflow_step_data["datasetid"] = workunit_definition.execution.dataset

    # Find or create the workflow step entity
    _create_workflow_step_if_not_exists(
        workflow=workflow, workflow_step_data=workflow_step_data, workunit_id=workunit_id, client=client
    )


def _create_workflow_step_if_not_exists(
    workflow: dict[str, Any], workflow_step_data: dict[str, Any], workunit_id: int, client: Bfabric
) -> None:
    workflow_step = WorkflowStep.find_by(workflow_step_data, client=client)
    if workflow_step:
        logger.info(f"Workflow step already exists: {workflow_step}, skipping creation.")
        return
    resp = client.save("workflowstep", workflow_step_data)
    logger.info(f"Created workflow step: {resp[0]['id']} for workunit {workunit_id} in workflow {workflow['id']}.")


def _find_or_create_workflow(
    workflow_template: WorkflowTemplate, workunit_definition: WorkunitDefinition, client: Bfabric
) -> dict[str, Any]:
    container_id = workunit_definition.registration.container_id
    workflow_data = {"containerid": container_id, "workflowtemplateid": workflow_template.id}
    resp = client.read("workflow", workflow_data)
    if len(resp) > 0:
        workflow = resp[0]
    else:
        resp = client.save("workflow", workflow_data)
        workflow = resp[0]
    return workflow


def _validate_chunks_list(work_dir: Path, chunk: str | None) -> list[Path]:
    """Converts the chunk argument to a list of Path objects.

    If it is None, all chunks from the chunks.yml file are returned.
    If it is specified, it is converted to a list with one element.
    All paths will be resolved to the work_dir and into absolute paths.
    """
    work_dir = work_dir.resolve()
    if chunk is None:
        return [work_dir / p for p in ChunksFile.read(work_dir=work_dir).chunks]
    else:
        return [work_dir / chunk]



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, assert_never

from bfabric_app_runner.inputs.resolve.resolved_inputs import (
    ResolvedInput,
    ResolvedFile,
    ResolvedStaticFile,
    ResolvedDirectory,
)
from bfabric_app_runner.util.checksums import md5sum

if TYPE_CHECKING:
    from pathlib import Path
    from bfabric.bfabric import Bfabric


class IntegrityState(Enum):
    """Represents the integrity state of a local artifact (file or directory).

    This enum is used to determine whether local artifacts need to be refreshed
    (re-downloaded or re-prepared) based on their current state and validation.

    States:
        Missing: The local artifact does not exist at the expected path.
                 Action: Download/prepare the artifact.

        NotChecked: The artifact exists but integrity validation was not performed
                   or is not possible (e.g., no checksum available).
                   Action: Assume the artifact is valid (no refresh needed).

        Correct: The artifact exists and passes all integrity checks
                (checksum validation, content validation, etc.).
                Action: No refresh needed.

        Incorrect: The artifact exists but fails integrity validation
                  (wrong checksum, corrupted content, etc.).
                  Action: Re-download/re-prepare the artifact.
    """

    Missing = "Missing"
    NotChecked = "NotChecked"
    Correct = "Correct"
    Incorrect = "Incorrect"

    def exists(self) -> bool:
        return self != IntegrityState.Missing


def check_integrity(file: ResolvedInput, local_path: Path, client: Bfabric) -> IntegrityState:
    """Checks the integrity of a local file/directory against the spec.

    The purpose of this integrity check is to determine whether the local artifact
    needs to be refreshed (re-downloaded/re-prepared). It validates:
    - File existence and basic accessibility
    - Checksum validation for files and archives
    - Content validation for static files
    - Basic structure validation for extracted directories

    This is NOT about validating filtering logic (include/exclude patterns) -
    that's handled during the preparation phase.
    """
    if not local_path.exists():
        return IntegrityState.Missing

    try:
        if isinstance(file, ResolvedFile):
            return _check_file_integrity(file, local_path)
        elif isinstance(file, ResolvedStaticFile):
            return _check_static_file_integrity(file, local_path)
        elif isinstance(file, ResolvedDirectory):
            return _check_directory_integrity(file, local_path)
        else:
            assert_never(file)
    except OSError:
        # File exists but can't be read - treat as incorrect to trigger refresh
        return IntegrityState.Incorrect


def _check_file_integrity(file: ResolvedFile, local_path: Path) -> IntegrityState:
    """Check integrity of a regular file."""
    if file.checksum is None:
        return IntegrityState.NotChecked
    else:
        return IntegrityState.Correct if file.checksum == md5sum(local_path) else IntegrityState.Incorrect


def _check_static_file_integrity(file: ResolvedStaticFile, local_path: Path) -> IntegrityState:
    """Check integrity of a static file by comparing content."""
    bytes_flag = "b" if isinstance(file.content, bytes) else ""
    with local_path.open(f"r{bytes_flag}") as f:
        return IntegrityState.Correct if f.read() == file.content else IntegrityState.Incorrect


def _check_directory_integrity(file: ResolvedDirectory, local_path: Path) -> IntegrityState:
    """Check integrity of an extracted directory and its cache file.

    For directories, we validate:
    1. The extracted directory exists and contains files
    2. The cache .zip file (if it exists) has correct checksum

    If no checksum is available, we do basic structural validation.
    """
    # Check if the extracted directory exists and is actually a directory
    if not local_path.is_dir():
        return IntegrityState.Incorrect

    # Basic structure validation - directory should contain files
    try:
        if not any(local_path.iterdir()):
            return IntegrityState.Incorrect
    except PermissionError:
        return IntegrityState.Incorrect

    # Check the cache .zip file if checksum is available
    if file.checksum is not None:
        cache_file = local_path.parent / f"{file.filename}.zip"
        if cache_file.exists():
            if not cache_file.is_file():
                return IntegrityState.Incorrect
            # Validate cache file checksum
            return IntegrityState.Correct if file.checksum == md5sum(cache_file) else IntegrityState.Incorrect
        else:
            # Cache file missing but we have checksum - trigger refresh
            return IntegrityState.Incorrect

    # No checksum available - can only do basic validation
    return IntegrityState.NotChecked



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import yaml

from bfabric_app_runner.commands.execute import execute_command
from bfabric_app_runner.inputs.prepare.prepare_folder import prepare_folder
from bfabric_app_runner.output_registration import register_outputs
from loguru import logger
from pydantic import BaseModel

from bfabric.experimental.workunit_definition import WorkunitDefinition

if TYPE_CHECKING:
    from bfabric_app_runner.specs.app.app_version import AppVersion
    from bfabric import Bfabric


class Runner:
    def __init__(self, spec: AppVersion, client: Bfabric, ssh_user: str | None = None) -> None:
        self._app_version = spec
        self._client = client
        self._ssh_user = ssh_user

    def run_dispatch(self, workunit_ref: int | Path, work_dir: Path) -> None:
        logger.info(f"Calling dispatch for workunit {workunit_ref} in {work_dir}")
        execute_command(self._app_version.commands.dispatch, str(workunit_ref), str(work_dir))

    def run_inputs(self, chunk_dir: Path) -> None:
        prepare_folder(
            inputs_yaml=chunk_dir / "inputs.yml",
            target_folder=chunk_dir,
            client=self._client,
            ssh_user=self._ssh_user,
            filter=None,
        )

    def run_process(self, chunk_dir: Path) -> None:
        logger.info(f"Calling process for chunk directory {chunk_dir}")
        execute_command(self._app_version.commands.process, str(chunk_dir))

    def run_collect(self, workunit_ref: int | Path, chunk_dir: Path) -> None:
        if self._app_version.commands.collect is not None:
            logger.info(f"Calling collect for workunit {workunit_ref} in {chunk_dir}")
            execute_command(self._app_version.commands.collect, str(workunit_ref), str(chunk_dir))
        else:
            logger.info("App does not have a collect step.")


class ChunksFile(BaseModel):
    # TODO move to better location
    chunks: list[Path]

    @classmethod
    def read(cls, work_dir: Path) -> ChunksFile:
        """Reads the chunks.yml file from the specified work directory."""
        return ChunksFile.model_validate(yaml.safe_load((work_dir / "chunks.yml").read_text()))


def run_app(
    app_spec: AppVersion,
    workunit_ref: int | Path,
    work_dir: Path,
    client: Bfabric,
    force_storage: Path | None,
    ssh_user: str | None = None,
    read_only: bool = False,
    dispatch_active: bool = True,
) -> None:
    """Executes all steps of the provided app."""
    # TODO would it be possible, to reuse the individual steps commands so there is certainly only one definition?
    work_dir = work_dir.resolve()
    workunit_ref = workunit_ref.resolve() if isinstance(workunit_ref, Path) else workunit_ref

    workunit_definition_file = work_dir / "workunit_definition.yml"
    workunit_definition = WorkunitDefinition.from_ref(
        workunit=workunit_ref, client=client, cache_file=workunit_definition_file
    )
    if not read_only:
        # Set the workunit status to processing
        client.save("workunit", {"id": workunit_definition.registration.workunit_id, "status": "processing"})

    runner = Runner(spec=app_spec, client=client, ssh_user=ssh_user)
    if dispatch_active:
        runner.run_dispatch(workunit_ref=workunit_definition_file, work_dir=work_dir)
    chunks_file = ChunksFile.read(work_dir=work_dir)
    for chunk in chunks_file.chunks:
        logger.info(f"Processing chunk {chunk}")
        runner.run_inputs(chunk_dir=chunk)
        runner.run_process(chunk_dir=chunk)
        runner.run_collect(workunit_ref=workunit_definition_file, chunk_dir=chunk)
        if not read_only:
            register_outputs(
                outputs_yaml=chunk / "outputs.yml",
                workunit_definition=workunit_definition,
                client=client,
                ssh_user=ssh_user,
                reuse_default_resource=app_spec.reuse_default_resource,
                force_storage=force_storage,
            )

    if not read_only:
        # Set the workunit status to available
        client.save("workunit", {"id": workunit_definition.registration.workunit_id, "status": "available"})



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import mako.template
from loguru import logger

if TYPE_CHECKING:
    from bfabric_app_runner.bfabric_integration.submitter.config.slurm_params import SlurmParameters


class SlurmJobTemplate:
    def __init__(self, params: SlurmParameters, workunit_id: int, wrapped_script: str, path: Path) -> None:
        self._params = params
        self._workunit_id = workunit_id
        self._wrapped_script = wrapped_script
        self._path = path

    @classmethod
    def default_path(cls) -> Path:
        return Path(__file__).parent / "slurm_job_template.bash.mako"

    def render_string(self) -> str:
        params = {
            "sbatch_params": self._params.sbatch_params,
            "wrapped_script": self._wrapped_script,
            "workunit_id": self._workunit_id,
        }
        logger.debug("Rendering {} with params: {}", self._path, params)
        template = mako.template.Template(filename=str(self._path))
        return template.render(**params)



from pathlib import Path
import shlex
import os
from bfabric_app_runner.commands.command_exec import execute_command_exec
from bfabric_app_runner.specs.app.commands_spec import CommandExec, CommandDocker, MountOptions


def _collect_mount_options(options: MountOptions, work_dir: Path) -> list[tuple[Path, Path, bool]]:
    """Collects all mounts that are required to run the command.

    These are returned as triplets of (source, target, read_only).
    """
    mounts = []
    if options.share_bfabric_config:
        mounts.append((Path("~/.bfabricpy.yml"), Path("/home/user/.bfabricpy.yml"), True))
    # TODO reconsider if we ever want work_dir_target to be customizable to be different from host path
    #      (currently things will break down if this is configured)
    work_dir_target = work_dir if options.work_dir_target is None else options.work_dir_target
    mounts.append((work_dir, work_dir_target, False))
    for source, target in options.read_only:
        mounts.append((source, target, True))
    for source, target in options.writeable:
        mounts.append((source, target, False))
    return [(source.expanduser().absolute(), target, read_only) for source, target, read_only in mounts]


def _to_shell(command: CommandDocker, work_dir: Path | None = None) -> list[str]:
    """Returns a shell command that can be used to run the specified command."""
    work_dir = (work_dir or Path()).expanduser().absolute()
    mounts = _collect_mount_options(command.mounts, work_dir)
    mount_args = []
    for host, container, read_only in mounts:
        source = shlex.quote(str(host))
        target = shlex.quote(str(container))
        mount_args.append("--mount")
        mount_args.append(f"type=bind,source={source},target={target}" + (",readonly" if read_only else ""))
    entrypoint_arg = ["--entrypoint", command.entrypoint] if command.entrypoint else []
    env_args = []
    for key, value in command.env.items():
        env_args.append("--env")
        env_args.append(f"{key}={shlex.quote(value)}")
    mac_address_arg = ["--mac-address", command.mac_address] if command.mac_address else []
    hostname_arg = ["--hostname", command.hostname] if command.hostname else []

    return [
        command.engine,
        "run",
        "--user",
        f"{os.getuid()}:{os.getgid()}",
        "--rm",
        *mount_args,
        *entrypoint_arg,
        *env_args,
        *mac_address_arg,
        *command.custom_args,
        *hostname_arg,
        command.image,
        *shlex.split(command.command),
    ]


def execute_command_docker(command: CommandDocker, *args: str) -> None:
    """Executes the command with the provided arguments."""
    shell_command = _to_shell(command)
    execute_command_exec(CommandExec(command=shlex.join(shell_command)), *args)



from pathlib import Path
from typing import Literal, Annotated

from pydantic import BaseModel, Discriminator, ConfigDict


class CommandShell(BaseModel):
    model_config = ConfigDict(extra="forbid")

    # TODO: Now deprecated by exec type - we can add a "bash" type later, if needed
    type: Literal["shell"] = "shell"
    """Identifies the command type."""

    command: str
    """The command to run, will be split by spaces and is not an actual shell script."""


class CommandExec(BaseModel):
    model_config = ConfigDict(extra="forbid")

    type: Literal["exec"] = "exec"
    """Identifies the command type."""

    command: str
    """The command to run, will be split by `shlex.split` and is not an actual shell script."""

    env: dict[str, str] = {}
    """Environment variables to set before executing the command."""

    prepend_paths: list[Path] = []
    """A list of paths to prepend to the PATH variable before executing the command.

    If multiple paths are specified, the first one will be the first in PATH, etc.
    """


class MountOptions(BaseModel):
    model_config = ConfigDict(extra="forbid")
    work_dir_target: Path | None = None
    read_only: list[tuple[Path, Path]] = []
    writeable: list[tuple[Path, Path]] = []
    share_bfabric_config: bool = True


class CommandDocker(BaseModel):
    model_config = ConfigDict(extra="forbid")
    type: Literal["docker"] = "docker"
    """Identifies the command type."""
    image: str
    """The container image to run."""
    command: str
    """The command to execute in the container."""
    entrypoint: str | None = None
    """The entrypoint to use for the container (instead of the image's default)."""
    engine: Literal["docker", "podman"] = "docker"
    """The container engine to use."""
    env: dict[str, str] = {}
    """Environment variables to set in the container."""
    mac_address: str | None = None
    """The MAC address to use for the container (instead of Docker's default assignment)."""
    mounts: MountOptions = MountOptions()
    """Mount options for the container."""
    hostname: str | None = None
    """The hostname to use for the container (instead of Docker's default assignment)."""
    custom_args: list[str] = []
    """Any custom CLI arguments to pass to the container engine."""


class CommandPythonEnv(BaseModel):
    model_config = ConfigDict(extra="forbid")

    type: Literal["python_env"] = "python_env"

    pylock: Path
    """Path to the Pylock file that specifies the environment to use."""

    command: str
    """The command to run, will be split by `shlex.split` and is not an actual shell script."""

    python_version: str | None = None
    """The Python version to use."""

    local_extra_deps: list[Path] = []
    """Additional dependencies to install into the environment.

    Each entry should be a path to a wheel, sdist, or local package directory.
    These will be installed into the environment with `uv pip install --no-deps` after the main requirements.
    No dependency resolution will be performed for these, so their dependencies should already be present
    in the environment (typically specified in the pylock file).
    """

    env: dict[str, str] = {}
    """Environment variables to set before executing the command."""

    prepend_paths: list[Path] = []
    """A list of paths to prepend to the PATH variable before executing the command.

    If multiple paths are specified, the first one will be the first in PATH, etc.
    """

    refresh: bool = False
    """When True, forces provisioning of an ephemeral environment that doesn't affect the cache.

    This creates a temporary environment for the single execution and doesn't modify or reuse
    the cached environment. This ensures that refresh operations always start from a clean state
    and don't leave broken environments in the cache if provisioning fails.
    """


Command = Annotated[CommandShell | CommandExec | CommandDocker | CommandPythonEnv, Discriminator("type")]


class CommandsSpec(BaseModel):
    """Defines the commands that are required to execute an app."""

    model_config = ConfigDict(extra="forbid", populate_by_name=True)

    dispatch: Command
    """The app dispatch command.

    It will be called with arguments: `$workunit_ref` `$work_dir`.
    """

    process: Command
    """The app process command.

    It will be called with arguments: `$chunk_dir`.
    """

    collect: Command | None = None
    """The app collect command, can be omitted if your process command already creates an `outputs.yml` file.

    It will be called with arguments: `$workunit_ref` `$chunk_dir`.
    """



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, assert_never

from loguru import logger

from bfabric_app_runner.inputs.prepare.prepare_resolved_directory import prepare_resolved_directory
from bfabric_app_runner.inputs.prepare.prepare_resolved_file import prepare_resolved_file
from bfabric_app_runner.inputs.prepare.prepare_resolved_static_file import prepare_resolved_static_file
from bfabric_app_runner.inputs.resolve.resolved_inputs import (
    ResolvedInputs,
    ResolvedFile,
    ResolvedStaticFile,
    ResolvedDirectory,
)
from bfabric_app_runner.inputs.resolve.resolver import Resolver
from bfabric_app_runner.specs.inputs_spec import (
    InputsSpec,
)

if TYPE_CHECKING:
    from pathlib import Path
    from bfabric.bfabric import Bfabric


def prepare_folder(
    inputs_yaml: Path,
    target_folder: Path | None,
    client: Bfabric,
    ssh_user: str | None,
    filter: str | None,
    action: Literal["prepare", "clean"] = "prepare",
) -> None:
    """Prepares the input files of a chunk folder according to the provided specs.

    :param inputs_yaml: Path to the inputs.yml file.
    :param target_folder: Path to the target folder where the input files should be downloaded.
    :param client: Bfabric client to use for obtaining metadata about the input files.
    :param ssh_user: SSH user to use for downloading the input files, should it be different from the current user.
    :param filter: only this input file will be prepared.
    :param action: Action to perform.
    """
    # set defaults
    inputs_yaml = inputs_yaml.absolute()
    if target_folder is None:
        target_folder = inputs_yaml.parent

    # parse the specs
    inputs_spec = InputsSpec.read_yaml(inputs_yaml)

    # resolve the specs
    resolver = Resolver(client=client)
    input_files = resolver.resolve(specs=inputs_spec.inputs)

    # Note: In some cases, this will be very inefficient to do after resolving compared to doing it before, but in
    #       general not all filenames will be known upfront.
    if filter is not None:
        input_files = input_files.apply_filter(filter_files=[filter])
        if not input_files.files:
            raise ValueError(f"Filter {filter} did not match any input files")

    # prepare the folder
    if action == "prepare":
        _prepare_input_files(input_files=input_files, working_dir=target_folder, ssh_user=ssh_user)
    elif action == "clean":
        _clean_input_files(input_files=input_files, working_dir=target_folder)
    else:
        raise ValueError(f"Unknown action: {action}")


def _prepare_input_files(input_files: ResolvedInputs, working_dir: Path, ssh_user: str | None) -> None:
    """Prepares the input files in the working directory according to the provided specs."""
    for input_file in input_files.files:
        match input_file:
            case ResolvedFile():
                prepare_resolved_file(file=input_file, working_dir=working_dir, ssh_user=ssh_user)
            case ResolvedStaticFile():
                prepare_resolved_static_file(file=input_file, working_dir=working_dir)
            case ResolvedDirectory():
                prepare_resolved_directory(file=input_file, working_dir=working_dir, ssh_user=ssh_user)
            case _:
                assert_never(input_file)


def _clean_input_files(input_files: ResolvedInputs, working_dir: Path) -> None:
    """Removes the specified files from working_dir, if they exist."""
    for input_file in input_files.files:
        path = working_dir / input_file.filename
        if path.exists():
            path.unlink()
            logger.info(f"Removed {path}")



from typing import assert_never

from bfabric_app_runner.commands.command_docker import execute_command_docker
from bfabric_app_runner.commands.command_exec import execute_command_exec
from bfabric_app_runner.commands.command_python_env import execute_command_python_env
from bfabric_app_runner.commands.command_shell import execute_command_shell
from bfabric_app_runner.specs.app.commands_spec import (
    Command,
    CommandShell,
    CommandExec,
    CommandDocker,
    CommandPythonEnv,
)


def execute_command(command: Command, *args: str) -> None:
    """Executes the given command with the provided arguments."""
    command_type = type(command)
    if issubclass(command_type, CommandShell):
        execute_command_shell(command, *args)
    elif issubclass(command_type, CommandExec):
        execute_command_exec(command, *args)
    elif issubclass(command_type, CommandDocker):
        execute_command_docker(command, *args)
    elif issubclass(command_type, CommandPythonEnv):
        execute_command_python_env(command, *args)
    else:
        assert_never(command_type)



from __future__ import annotations

from typing import Literal

from pydantic import BaseModel, ConfigDict

from bfabric_app_runner.specs.common_types import RelativeFilePath  # noqa: TC001


class BfabricResourceSpec(BaseModel):
    model_config = ConfigDict(extra="forbid")
    type: Literal["bfabric_resource"] = "bfabric_resource"

    id: int
    """B-Fabric resource ID"""

    filename: RelativeFilePath | None = None
    """Target filename to save to"""

    check_checksum: bool = True
    """Whether to check the checksum of the file, after downloading"""



from __future__ import annotations
from typing import TYPE_CHECKING, Literal

from pydantic import BaseModel

if TYPE_CHECKING:
    from bfabric import Bfabric


class StaticFileSpec(BaseModel):
    type: Literal["static_file"] = "static_file"

    content: str | bytes
    """The text or binary content to write."""
    filename: str
    """The target filename to write to."""

    def resolve_filename(self, client: Bfabric) -> str:
        return self.filename



from __future__ import annotations

from typing import Annotated

from pydantic import Field

AbsoluteFilePath = Annotated[str, Field(pattern=r"^/[^:]*$")]
"""Absolute file path, excluding ":" characters."""

RelativeFilePath = Annotated[str, Field(pattern=r"^[^/][^:]*$")]
"""Relative file path, excluding absolute paths and ":" characters."""



from __future__ import annotations
from enum import Enum
from pathlib import Path  # noqa
from typing import Literal, Annotated

from pydantic import Field

from bfabric_app_runner.actions.config_file import FromConfigFile


class Action(str, Enum):
    run = "run"
    dispatch = "dispatch"
    inputs = "inputs"
    process = "process"
    outputs = "outputs"


class ActionDispatch(FromConfigFile):
    action: Literal[Action.dispatch] = Action.dispatch
    work_dir: Path
    app_ref: Path
    workunit_ref: int | Path
    read_only: bool = False


class ActionRun(FromConfigFile):
    action: Literal[Action.run] = Action.run
    work_dir: Path
    chunk: str | None = None
    ssh_user: str | None = None
    filter: str | None = None
    app_ref: Path
    force_storage: Path | None = None
    read_only: bool = False
    workunit_ref: int | Path


class ActionInputs(FromConfigFile):
    action: Literal[Action.inputs] = Action.inputs
    work_dir: Path
    chunk: str | None = None
    ssh_user: str | None = None
    filter: str | None = None

    @classmethod
    def from_action_run(cls, action_run: ActionRun, chunk: str | None) -> ActionInputs:
        """Returns an ActionInputs object from an ActionRun object."""
        return ActionInputs(
            work_dir=action_run.work_dir,
            chunk=action_run.chunk if chunk is None else chunk,
            ssh_user=action_run.ssh_user,
            filter=action_run.filter,
        )


class ActionProcess(FromConfigFile):
    action: Literal[Action.process] = Action.process
    work_dir: Path
    app_ref: Path
    chunk: str | None = None

    @classmethod
    def from_action_run(cls, action_run: ActionRun, chunk: str | None) -> ActionProcess:
        """Returns an ActionProcess object from an ActionRun object."""
        return ActionProcess(
            work_dir=action_run.work_dir,
            app_ref=action_run.app_ref,
            chunk=action_run.chunk if chunk is None else chunk,
        )


class ActionOutputs(FromConfigFile):
    action: Literal[Action.outputs] = Action.outputs
    work_dir: Path
    workunit_ref: int | Path
    app_ref: Path
    chunk: str | None = None
    ssh_user: str | None = None
    force_storage: Path | None = None
    read_only: bool = False

    @classmethod
    def from_action_run(cls, action_run: ActionRun, chunk: str | None) -> ActionOutputs:
        """Returns an ActionOutputs object from an ActionRun object."""
        return ActionOutputs(
            work_dir=action_run.work_dir,
            app_ref=action_run.app_ref,
            chunk=action_run.chunk if chunk is None else chunk,
            ssh_user=action_run.ssh_user,
            force_storage=action_run.force_storage,
            read_only=action_run.read_only,
            workunit_ref=action_run.workunit_ref,
        )


ActionGeneric = Annotated[
    ActionRun | ActionDispatch | ActionInputs | ActionProcess | ActionOutputs, Field(discriminator="action")
]



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from bfabric_app_runner.inputs.list_inputs.integrity import check_integrity, IntegrityState
from bfabric_app_runner.inputs.resolve.resolver import Resolver
from rich.console import Console
from rich.table import Table, Column

if TYPE_CHECKING:
    from bfabric_app_runner.specs.inputs_spec import InputSpecType
    from pathlib import Path
    from bfabric.bfabric import Bfabric


@dataclass
class FileState:
    name: str
    path: Path
    exists: bool
    integrity: IntegrityState


def list_input_states(
    specs: list[InputSpecType],
    target_folder: Path,
    client: Bfabric,
    check_files: bool,
) -> list[FileState]:
    """Returns the states of the input files, performing integrity checks if requested.

    :param specs: List of input specs to consider
    :param target_folder: The target folder where the files should be stored
    :param client: A B-Fabric client
    :param check_files: If this is `True`, in addition to listing the files, their integrity states will be computed,
        otherwise the value is always `NotChecked`.
    """
    input_states = []
    resolver = Resolver(client=client)
    input_files = resolver.resolve(specs=specs)

    for input_file in input_files.files:
        path = target_folder / input_file.filename
        exists = path.exists()
        if not check_files:
            integrity = IntegrityState.NotChecked
        else:
            integrity = check_integrity(file=input_file, local_path=path, client=client)
        input_states.append(FileState(name=input_file.filename, path=path, exists=exists, integrity=integrity))
    return input_states


def print_input_states(input_states: list[FileState]) -> None:
    """Prints the states of the input files to the command line."""
    table = Table(
        Column("File"),
        Column("Exists"),
        Column("Integrity"),
    )
    for state in input_states:
        table.add_row(
            str(state.name),
            {True: "Yes", False: "No"}[state.exists],
            state.integrity.value,
        )
    console = Console()
    console.print(table)
